/*
 * Copyright (c) 2009,
 * @author Mitsuharu Morisawa
 *
 * AIST
 *
 * All rights reserved.
 *
 * This program is made available under the terms of the Eclipse Public License
 * v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 */
#ifndef __MWALK_H__
#define __MWALK_H__

#include "Plugin_impl.h"
#include "SequencePlayer.h"
#include "corba_util.h"
#include "MwalkGenerator.h"
#include "MwalkFootPlanner.h"
#include "MeasureExecTime.h"
#include "StateLogger.h"
#include "ReadText.h"

using namespace foot_planner;


namespace mwalk_pattern_generator{
  
#define robot_state   RobotState
#define motor_command RobotState
  
#define CHECK_EXEC_TIME
  
#ifdef CHECK_EXEC_TIME
#define clearExecTime(a) a.clearExecTime()
#define initExecTime(a)  a.initExecTime()
#define pushExecTime(a)  a.pushExecTime()
#define calcExecTime(a)  a.calcExecTime()
#else
#define clearExecTime(a)
#define initExecTime(a)
#define pushExecTime(a)
#define calcExecTime(a)
#endif
  
  class mwalk : public plugin{
  private:
    bool wait_request;
    bool state_logger_mode;
    
    MwalkGenerator *pg;
    MwalkFootPlanner *fp;
    SequencePlayer_var seq;
    
    MeasureExecTime mExecTime;
    StateLogger state_logger;
    friend class mwalk_state;     
    
  protected:
    void setupMethodTable(void);
    void pushStateLogger(void);
    int  getFloor(string& str_floor);
    
  public:
    mwalk(istringstream& strm);
    ~mwalk();
    
    bool setup(robot_state *rs, motor_command *mc);
    void control(robot_state *rs, motor_command *mc);
    bool cleanup(robot_state *rs, motor_command *mc);
    
    plugin_state *get_state();
    
    void m_start_walk(istringstream& strm);
    void m_start_walk_nowait(istringstream& strm);
    void m_start_zmpfix(istringstream& strm);
    void m_stop_zmpfix(istringstream& strm);
    void m_set_foot_pos(istringstream& strm);
    void m_set_next_foot_pos(istringstream& strm);
    void m_set_target_pos(istringstream& strm);
    void m_set_target_pos_nowait(istringstream& strm);
    void m_set_target_pos_passgate(istringstream& strm);
    void m_set_arc(istringstream& strm);
    void m_set_arc_nowait(istringstream& strm);
    void m_push_walk_command(istringstream& strm);
    void m_first_step(istringstream& strm);
    
    void m_set_profile_horizontal(istringstream& strm);
    void m_set_profile_vertical(istringstream &strm);
    void m_set_pivot_translation(istringstream& strm);
    void m_set_ankle_rotation(istringstream& strm);
    
    void m_set_waist_middle_ratio(istringstream& strm);
    void m_set_swing_height(istringstream& strm);
    void m_set_landoff_angle(istringstream& strm);
    void m_set_touchdown_angle(istringstream& strm);
    void m_set_backstep_pitch_angle(istringstream& strm);
    void m_set_step_limit_x(istringstream& strm);
    void m_set_step_limit_y(istringstream& strm);
    void m_set_step_limit_z(istringstream& strm);
    void m_set_step_limit_yaw(istringstream& strm);
    void m_set_step_yaw_ratio(istringstream& strm);
    void m_set_waist_yaw_ratio(istringstream& strm);
    void m_set_arm_swing_ratio(istringstream& strm);
    void m_set_zmp_ankle(istringstream& strm);
    void m_set_zmp_path(istringstream& strm);
    void m_set_ankle_ds_rotation_ratio(istringstream& strm);
    void m_trunk_roll_mode(istringstream& strm);
    
    void m_time_adj(istringstream& strm);
    void m_load_footpos(istringstream& strm);
    void m_zmpshaping(istringstream& strm);
    void m_dfilter_preview(istringstream& strm);
    void m_init_cog_offset_gain(istringstream& strm);
    void m_update_cog_offset_gain(istringstream& strm);
    void m_load_zmpshaping_pgain(istringstream& strm);
    void m_load_dfilter_pgain(istringstream& strm);
    void m_reactive_motion(istringstream& strm);
    void m_state_logger_mode(istringstream& strm);
    void m_save_state_logger(istringstream& strm);
    void m_stat(istringstream& strm);
    void m_check_joints_status(istringstream& strm);
    void m_joints_status(istringstream& strm);
    void m_debug_level(istringstream& strm);
    void m_debug_info(istringstream& strm);
    
#ifdef WITH_TOE_JOINT
    void m_set_zmp_toe(istringstream& strm);
    void m_set_toe_limit_pitch(istringstream& strm);
#endif
  };
  
  class mwalk_state : public plugin_state{
    
  private:
    int n_clock;
    WalkingState ws;
    WalkingState ws_pre;
    vector3 Zmp_cmd;
    vector3 Zmp_shp;
    vector3 Zmp_ref;
    vector3 VZmp_cmd;
    vector3 Zmp_cmd_pre;
    vector3 Zmp_model;
    vector3 Zmp_ideal;
    vector3 Zmp_model2;
    vector3 Zmp_max;
    vector3 Zmp_min;
    vector3 Gsensor_offset;
    vector3 Xg_cmd;
    vector3 Xg_shp;
    vector3 Xg_dfilter;
    vector3 Vg_cmd;
    vector3 Vgdot_cmd;
    vector3 Xf_cmd[2];
    vector3 Vf_cmd[2];
    vector3 Xa_cmd[2];
    vector3 Va_cmd[2];
    vector3 Ra_cmd[2];
    vector3 Wa_cmd[2];
    vector3 Xwaist_cmd;
    vector3 Vwaist_cmd;
    vector3 Rwaist_cmd;
    vector3 Wwaist_cmd;
    vector3 Pivot_cmd[2];
    
    vector3 VCOGdot_res;
    vector3 Zmp_res;
    vector3 COG_cmp;
    vector3 VCOG_cmp;
    
    vector3 P_cmd;
    vector3 L_cmd;
    vector3 P_model;
    vector3 L_model;
    
    vector3 Pvertex[4][2];
    vector3 Vvertex[4][2];
    
    Jvector q_cmd;
    Jvector qdot_cmd;
    
  public:
    mwalk_state(mwalk *mg);
    ~mwalk_state();
    const char *header();
    void pprint(ostream &s) const;
    
  };

} // end of namespace mwalk_pattern_generator

#endif // __MWALK_H__
