/*
 * Copyright (c) 2009,
 * @author Mitsuharu Morisawa
 *
 * AIST
 *
 * All rights reserved.
 *
 * This program is made available under the terms of the Eclipse Public License
 * v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 */

#include "AnalyticalSolution.h"

using namespace mwalk_pattern_generator;


AnalyticalSolution::AnalyticalSolution(void)
{
  noalias(A) = dzeromatrix(Ncog*2+6, Ncog*2+6);
}

AnalyticalSolution::~AnalyticalSolution()
{
}

void AnalyticalSolution::getMotion(double k_[][3], vector3& p_cmd, vector3& v_cmd, vector3& a_cmd, vector3& zmp_cmd, vector3& vzmp_cmd, double *tm_, double tn, int n)
{
  int ns = 0;
  double t_total = tm_[0];
  while( ns < n-1 && t_total + tm_[ns+1] < tn ){
    t_total += tm_[ns+1];
    ns++;
  }
  
  if( ns == n-1 && tn >= t_total + tm_[n] )
    tn = t_total + tm_[n];

  double ts  = tn - t_total;
  double ts2 = ts * ts;
  double ts3 = ts * ts2;
  double ts4 = ts * ts3;
  double ts5 = ts * ts4;

  double e = exp(omega[ns]*ts);
  double e_inv = 1.0 / e;
  double ch = (e + e_inv) / 2.0;
  double sh = (e - e_inv) / 2.0;

  double A_coef[5];
  for( int i = X ; i < Z ; i++ ){
    for( int j = 1 ; j <= 5 ; j++ ){
      A_coef[j-1] = k_[ns*7+j+1][i];
      double cf = 1.0;
      for( int k = j ; k <= 3 ; k+=2 ){
	cf *= (double)(k*(k+1))/omega2[ns];
	A_coef[j-1] += cf * k_[ns*7+k+3][i];
      }
    }
    
    p_cmd(i) = k_[ns*7][i]*ch + k_[ns*7+1][i]*sh + A_coef[0] + A_coef[1]*ts + A_coef[2] * ts2 + A_coef[3]*ts3 + A_coef[4]*ts4;
    v_cmd(i) = omega[ns] *(k_[ns*7][i]*sh + k_[ns*7+1][i]*ch) + A_coef[1] + 2.0*A_coef[2]*ts + 3.0*A_coef[3]*ts2 + 4.0*A_coef[4]*ts3;
    a_cmd(i) = omega2[ns]*(k_[ns*7][i]*ch + k_[ns*7+1][i]*sh) + 2.0*A_coef[2] + 6.0*A_coef[3]*ts + 12.0*A_coef[4]*ts2;
    zmp_cmd(i)  = k_[ns*7+2][i] + k_[ns*7+3][i]*ts + k_[ns*7+4][i]*ts2 + k_[ns*7+5][i]*ts3 + k_[ns*7+6][i]*ts4;
    vzmp_cmd(i) = k_[ns*7+3][i] + 2.0*k_[ns*7+4][i]*ts + 3.0*k_[ns*7+5][i]*ts2 + 4.0*k_[ns*7+6][i]*ts3;
  }

  zmp_cmd(Z)  = k_[ns*7  ][Z] + k_[ns*7+1][Z]*ts + k_[ns*7+2][Z]*ts2 + k_[ns*7+3][Z]*ts3 + k_[ns*7+4][Z]*ts4 + k_[ns*7+5][Z]*ts5;
  vzmp_cmd(Z) = k_[ns*7+1][Z] + 2.0*k_[ns*7+2][Z]*ts + 3.0*k_[ns*7+3][Z]*ts2 + 4.0*k_[ns*7+4][Z]*ts3 + 5.0*k_[ns*7+5][Z]*ts4;
}

void AnalyticalSolution::getMotion(double k_[][3], vector3& p_cmd, vector3& v_cmd, vector3& zmp_cmd, vector3& vzmp_cmd, double *tm_, double tn, int n)
{
  vector3 a_cmd(0.0);
  getMotion(k_, p_cmd, v_cmd, a_cmd, zmp_cmd, vzmp_cmd, tm_, tn, n);  
}

// ZMPの単脚期の移動には未対応
void AnalyticalSolution::makeZMPError(dvector *zmp_err_, vector3 *zmp_des_, double kg_[][3], double *tm_)
{
  int Nmax = 0;
  for( int i = 1 ; i < Ncog ; i++ )
    Nmax += (int)round(tm_[i]/DT);

  if( Nmax > (int)zmp_err_[0].size() )
    Nmax = (int)zmp_err_[0].size();
  
  int ns = 0;
  double t_total = 0.0;
  for( int i = 0 ; i < Nmax ; i++ ){
    double tn = (double)i*DT;
    while( ns < Ncog-2 && t_total + tm_[ns+1] < tn ){
      t_total += tm_[ns+1];
      ns++;
    }
    
    if( ns == Ncog-2 && tn >= t_total + tm_[Ncog-1] )
      tn = t_total + tm_[Ncog-1];
      
    double ts  = tn - t_total;
    double ts2 = ts * ts;
    double ts3 = ts * ts2;
    double ts4 = ts * ts3;
    
    double zmp_ref[2];
    zmp_ref[0] = kg_[ns*7+2][0] + kg_[ns*7+3][0]*ts + kg_[ns*7+4][0]*ts2 + kg_[ns*7+5][0]*ts3 + kg_[ns*7+6][0]*ts4;
    zmp_ref[1] = kg_[ns*7+2][1] + kg_[ns*7+3][1]*ts + kg_[ns*7+4][1]*ts2 + kg_[ns*7+5][1]*ts3 + kg_[ns*7+6][1]*ts4;

    for( int j = X ; j <= Y ; j++ ){
      if( zmp_des_[ns](j) == zmp_des_[ns+1](j) )
	zmp_err_[j](i) = zmp_ref[j] - zmp_des_[ns](j);
      else
	zmp_err_[j](i) = 0.0;
    }
  }

  for( int i = Nmax ; i < (int)zmp_err_[0].size() ; i++ ){
    zmp_err_[0](i) = 0.0;
    zmp_err_[1](i) = 0.0;
  }
}

//
// Kg = [V1 W1 a2 a3 a4 V2 W2 ... Vn-1 Wn-1 Vn Wn a2 a3 a4]^T
// Kg = Aginv * Bg
//
// xgi_cmd_, vgi_cmd_ : initial COG conditions
// tm_[0] : current time
// zmp_des_[0] vzmp_des_[0] : initial ZMP conditions
// tm_[1]
// zmp_des_[1] vzmp_des_[1] : second ZMP conditions
// ...
// tm_[Ncog]
// zmp_des_[Ncog] vzmp_des_[Ncog] : terimal ZMP conditions
// 
void AnalyticalSolution::makeCoefParam(ANvector *b_, double *tm_, vector3 *zmp_des_, vector3 *vzmp_des_, vector3& xgi_cmd_, vector3& vgi_cmd_)
{
  double tm2_[Ncog], tm3_[Ncog];
  for( int i = 0 ; i < Ncog ; i++ ){
    tm2_[i] = tm_[i+1] * tm_[i+1];
    tm3_[i] = tm_[i+1] * tm2_[i];
  }

  // terminal COG condition
  vector3 Xgf_cmd( zmp_des_[Ncog] );
  vector3 Vgf_cmd( 0.0 );

  double zmp_coef[Ncog-2][4];
  for( int k = X ; k <= Y ; k++ ){
    b_[k](0) = xgi_cmd_(k) - zmp_des_[0](k);
    b_[k](1) = vgi_cmd_(k) - vzmp_des_[0](k);
    b_[k](2) = zmp_des_[1](k)  - zmp_des_[0](k) - vzmp_des_[0](k)*tm_[1];
    b_[k](3) = vzmp_des_[1](k) - vzmp_des_[0](k);
    
    for( int i = 2 ; i < Ncog ; i++ ){
      zmp_coef[i-2][0] = zmp_des_[i-1](k);
      zmp_coef[i-2][1] = vzmp_des_[i-1](k);
      zmp_coef[i-2][2] = ( 3.0*(zmp_des_[i](k) - zmp_des_[i-1](k)) - (vzmp_des_[i](k)+2.0*vzmp_des_[i-1](k))*tm_[i]) / tm2_[i-1];
      zmp_coef[i-2][3] = (-2.0*(zmp_des_[i](k) - zmp_des_[i-1](k)) + (vzmp_des_[i](k)+    vzmp_des_[i-1](k))*tm_[i]) / tm3_[i-1];
    }
    
    b_[k](4) = zmp_coef[0][0] + 2.0*zmp_coef[0][2]/omega2[1] - zmp_des_[0](k) - vzmp_des_[0](k)*tm_[1];
    b_[k](5) = zmp_coef[0][1] + 6.0*zmp_coef[0][3]/omega2[1] - vzmp_des_[0](k);
    
    for( int i = 2 ; i < Ncog-1 ; i++ ){
      b_[k](i*2+2) = zmp_coef[i-1][0] + 2.0*zmp_coef[i-1][2]/omega2[i] - zmp_coef[i-2][0]
	- tm_[i] * zmp_coef[i-2][1] - (2.0/omega2[i-1]+tm2_[i-1]) * zmp_coef[i-2][2] - (6.0*tm_[i]/omega2[i-1]+tm3_[i-1]) * zmp_coef[i-2][3];
      b_[k](i*2+3) = zmp_coef[i-1][1] + 6.0*zmp_coef[i-1][3]/omega2[i] - zmp_coef[i-2][1]
	- 2.0*tm_[i] * zmp_coef[i-2][2] - (6.0/omega2[i-1]+3.0*tm2_[i-1]) * zmp_coef[i-2][3];
    }
    
    b_[k](Ncog*2)   = zmp_des_[Ncog-1](k) - zmp_coef[Ncog-3][0] - tm_[Ncog-1]*zmp_coef[Ncog-3][1]
      - (2.0/omega2[Ncog-2]+tm2_[Ncog-2]) * zmp_coef[Ncog-3][2] - (6.0*tm_[Ncog-1]/omega2[Ncog-2]+tm3_[Ncog-2]) * zmp_coef[Ncog-3][3];
    b_[k](Ncog*2+1) = vzmp_des_[Ncog-1](k) - zmp_coef[Ncog-3][1] - 2.0*tm_[Ncog-1] * zmp_coef[Ncog-3][2]
      - (6.0/omega2[Ncog-2]+3.0*tm2_[Ncog-2]) * zmp_coef[Ncog-3][3];    
    
    b_[k](Ncog*2+2) = Xgf_cmd(k) - zmp_des_[Ncog-1](k) - vzmp_des_[Ncog-1](k) * tm_[Ncog];
    b_[k](Ncog*2+3) = Vgf_cmd(k) - vzmp_des_[Ncog-1](k);
    b_[k](Ncog*2+4) = zmp_des_[Ncog](k)  - zmp_des_[Ncog-1](k) - vzmp_des_[Ncog-1](k) * tm_[Ncog];
    b_[k](Ncog*2+5) = vzmp_des_[Ncog](k) - vzmp_des_[Ncog-1](k);
  }
}

void AnalyticalSolution::makeCoefMatrix(ANmatrix& ainv_, double *tm_)
{
  double ch[Ncog], sh[Ncog], tm2_[Ncog], tm3_[Ncog], tm4_[Ncog];
  for( int i = 0 ; i < Ncog ; i++ ){
    double e = exp(omega[i] * tm_[i+1]);
    double e_inv = 1.0 / e;
    ch[i] = (e + e_inv) / 2.0;
    sh[i] = (e - e_inv) / 2.0;

    tm2_[i] = tm_[i+1] * tm_[i+1];
    tm3_[i] = tm_[i+1] * tm2_[i];
    tm4_[i] = tm_[i+1] * tm3_[i];
  }
  
  // Inital condition (i = 1)
  A(0,0) = 1.0;        A(0,2) = 2.0/omega2[0];  A(0,4) = 24.0/omega4[0];
  A(1,1) = omega[0];   A(1,3) = 6.0/omega2[0];
  A(2,2) = tm2_[0];    A(2,3) = tm3_[0];     A(2,4) = tm4_[0];
  A(3,2) = 2.0*tm_[1]; A(3,3) = 3.0*tm2_[0]; A(3,4) = 4.0*tm3_[0];

  // Continuity condition (i = 2...Ncog-1)
  A(4,0) = ch[0];                        A(4,1) = sh[0];                                         A(4,2) = 2.0/omega2[0]+tm2_[0];
  A(4,3) = 6.0*tm_[1]/omega2[0]+tm3_[0]; A(4,4) = 24.0/omega4[0]+12.0*tm2_[0]/omega2[0]+tm4_[0]; A(4,5) = -1.0;
  A(5,0) = omega[0]*sh[0];               A(5,1) = omega[0]*ch[0];                                A(5,2) = 2.0*tm_[1];
  A(5,3) = 6.0/omega2[0]+3.0*tm2_[0];    A(5,4) = 24.0*tm_[1]/omega2[0]+4.0*tm3_[0];             A(5,6) = -omega[1];
   
  for( int i = 1 ; i <= Ncog-3 ; i++ ){
    A(i*2+4,i*2+3) = ch[i];          A(i*2+4,i*2+4) = sh[i];          A(i*2+4,i*2+5) = -1.0;
    A(i*2+5,i*2+3) = omega[i]*sh[i]; A(i*2+5,i*2+4) = omega[i]*ch[i]; A(i*2+5,i*2+6) = -omega[i+1];
  }

  A(Ncog*2  ,Ncog*2-1) =  ch[Ncog-2];               A(Ncog*2  ,Ncog*2  ) =  sh[Ncog-2];
  A(Ncog*2  ,Ncog*2+1) = -1.0;                      A(Ncog*2  ,Ncog*2+3) = -2.0/omega2[Ncog-1];       A(Ncog*2  ,Ncog*2+5) = -24.0/omega4[Ncog-1];
  A(Ncog*2+1,Ncog*2-1) =  omega[Ncog-2]*sh[Ncog-2]; A(Ncog*2+1,Ncog*2  ) =  omega[Ncog-2]*ch[Ncog-2];
  A(Ncog*2+1,Ncog*2+2) = -omega[Ncog-1];            A(Ncog*2+1,Ncog*2+4) = -6.0/omega2[Ncog-1];
  
  // Terminal condition (i = Ncog)
  A(Ncog*2+2,Ncog*2+1) = ch[Ncog-1];                      A(Ncog*2+2,Ncog*2+2) = sh[Ncog-1];
  A(Ncog*2+2,Ncog*2+3) = 2.0/omega2[Ncog-1]+tm2_[Ncog-1]; A(Ncog*2+2,Ncog*2+4) = 6.0*tm_[Ncog]/omega2[Ncog-1]+tm3_[Ncog-1];
  A(Ncog*2+2,Ncog*2+5) = 24.0/omega4[Ncog-1]+12.0*tm2_[Ncog-1]/omega2[Ncog-1]+tm4_[Ncog-1];
  A(Ncog*2+3,Ncog*2+1) = omega[Ncog-1]*sh[Ncog-1];        A(Ncog*2+3,Ncog*2+2) = omega[Ncog-1]*ch[Ncog-1];
  A(Ncog*2+3,Ncog*2+3) = 2.0*tm_[Ncog];                   A(Ncog*2+3,Ncog*2+4) = 6.0/omega2[Ncog-1]+3.0*tm2_[Ncog-1]; A(Ncog*2+3,Ncog*2+5) = 24.0*tm_[Ncog]/omega2[Ncog-1]+4.0*tm3_[Ncog-1];
  A(Ncog*2+4,Ncog*2+3) = tm2_[Ncog-1];                    A(Ncog*2+4,Ncog*2+4) = tm3_[Ncog-1];                        A(Ncog*2+4,Ncog*2+5) = tm4_[Ncog-1];
  A(Ncog*2+5,Ncog*2+3) = 2.0*tm_[Ncog];                   A(Ncog*2+5,Ncog*2+4) = 3.0*tm2_[Ncog-1];                    A(Ncog*2+5,Ncog*2+5) = 4.0*tm3_[Ncog-1];

  noalias(ainv_) = inverse(A);

  //cout << "A = " << A << endl;
  //cout << "ainv_ = " << ainv_ << endl;
}

void AnalyticalSolution::makeCoefVector(double k_[][3], ANmatrix& ainv_, ANvector *b_, double *tm_, vector3 *zmp_des_, vector3 *vzmp_des_)
{
  double tm2_[Ncog], tm3_[Ncog], tm4_[Ncog], tm5_[Ncog];
  for( int i = 0 ; i < Ncog ; i++ ){
    tm2_[i] = tm_[i+1] * tm_[i+1];
    tm3_[i] = tm_[i+1] * tm2_[i];
    tm4_[i] = tm_[i+1] * tm3_[i];
    tm5_[i] = tm_[i+1] * tm4_[i];
  }
  
  ANvector kd;
  for( int k = X ; k <= Y ; k++ ){
    noalias(kd) = prod(ainv_, b_[k]);

    k_[0][k] = kd(0);
    k_[1][k] = kd(1);
    k_[2][k] = zmp_des_[0](k);
    k_[3][k] = vzmp_des_[0](k);
    k_[4][k] = kd(2);
    k_[5][k] = kd(3);
    k_[6][k] = kd(4);

    for( int i = 1 ; i < Ncog-1 ; i++){
      k_[i*7  ][k] = kd(i*2+3);
      k_[i*7+1][k] = kd(i*2+4);
      k_[i*7+2][k] = zmp_des_[i](k);
      k_[i*7+3][k] = vzmp_des_[i](k);
      k_[i*7+4][k] = ( 3.0*(zmp_des_[i+1](k) - zmp_des_[i](k)) - (vzmp_des_[i+1](k)+2.0*vzmp_des_[i](k))*tm_[i+1]) / tm2_[i];
      k_[i*7+5][k] = (-2.0*(zmp_des_[i+1](k) - zmp_des_[i](k)) + (vzmp_des_[i+1](k)+    vzmp_des_[i](k))*tm_[i+1]) / tm3_[i];
      k_[i*7+6][k] = 0.0;
    }

    k_[Ncog*7-7][k] = kd(Ncog*2+1);
    k_[Ncog*7-6][k] = kd(Ncog*2+2);
    k_[Ncog*7-5][k] = zmp_des_[Ncog-1](k);
    k_[Ncog*7-4][k] = vzmp_des_[Ncog-1](k);
    k_[Ncog*7-3][k] = kd(Ncog*2+3);
    k_[Ncog*7-2][k] = kd(Ncog*2+4);
    k_[Ncog*7-1][k] = kd(Ncog*2+5);
  }

  for( int i = 0 ; i < Ncog ; i++ ){
    const double ai = 0.0, af = 0.0;
    k_[i*7  ][Z] = zmp_des_[i](Z);
    k_[i*7+1][Z] = vzmp_des_[i](Z);
    k_[i*7+2][Z] = ai / 2.0;
    k_[i*7+3][Z] = ( 20.0*(zmp_des_[i+1](Z)-zmp_des_[i](Z)) - ( 8.0*vzmp_des_[i+1](Z)+12.0*vzmp_des_[i](Z))*tm_[i+1] + (    af-3.0*ai)*tm2_[i]) / (2.0*tm3_[i]);
    k_[i*7+4][Z] = (-30.0*(zmp_des_[i+1](Z)-zmp_des_[i](Z)) + (14.0*vzmp_des_[i+1](Z)+16.0*vzmp_des_[i](Z))*tm_[i+1] - (2.0*af-3.0*ai)*tm2_[i]) / (2.0*tm4_[i]);
    k_[i*7+5][Z] = ( 12.0*(zmp_des_[i+1](Z)-zmp_des_[i](Z)) -  6.0*(vzmp_des_[i+1](Z)+vzmp_des_[i](Z))*tm_[i+1] + (af-ai)*tm2_[i]) / (2.0*tm5_[i]);
    k_[i*7+6][Z] = 0.0;
  }

  //cout << "Kd:" << Kd << endl;
  //cout << "Kg = " << Kg << endl;
}

/////////////////////////////////////////////////////////////////////////////
// fast calculation of the COG trajectory

void AnalyticalSolution::setCoefMatrix(dmatrix &ainv_, matrix22 *A_, matrix24 *B_, vector2 *C_, matrix22 *An_, matrix24 *Bn_, double *tm_, WalkingState ws_)
{
  for( int i = 0 ; i < Ncog ; i++ ){
    const double tm  = tm_[i+1];
    const double tm2 = tm * tm;
    const double tm3 = tm * tm2;
    const double tm4 = tm * tm3;
    
    const double w  = omega[i];
    const double w2 = omega2[i];
    const double w4 = omega4[i];
    const double e = exp(w * tm);
    const double ch = (e * e + 1.0) / (2.0 * e);
    const double sh = (e * e - 1.0) / (2.0 * e);
    
    A_[i] =
      ch,   sh/w,
      w*sh, ch;
    
    matrix24 B11;
    B11 =
      1.0-ch, tm-sh/w, 0.0, 0.0,
      -w*sh,  1.0-ch,  0.0, 0.0;
    matrix22 B12;
    B12 =
      2.0*(1.0-ch)/w2+tm2, 6.0*(tm-sh/w)/w2 + tm3,
      2.0*(tm-sh/w),       6.0*(1.0-ch)/w2 + 3.0*tm2;
    matrix24 B13;
    B13 =
      -3.0/tm2, -2.0/tm, 3.0/tm2,  -1.0/tm,
      2.0/tm3,  1.0/tm2, -2.0/tm3, 1.0/tm2;
    B_[i] = B11 + B12 * B13;

    vector2 C1;
    C1 =
      tm2,
      -2.0*tm;
    vector2 C2;
    C2 =
      24.0*(1.0-ch)/w4 + 12.0*tm2/w2 + tm4,
      24.0*(tm-sh/w)/w2 + 4.0*tm3;
    C_[i] = B12 * C1 + C2;
  }
  
  An_[Ncog-1] = A_[Ncog-1];
  for( int i = Ncog-2 ; i >= 0 ; i-- )
    An_[i] = An_[i+1] * A_[i];
  
  for( int i = 0 ; i < Ncog-1 ; i++ )
    Bn_[i] = An_[i+1] * B_[i];
  Bn_[Ncog-1] = B_[Ncog-1];
  
  vector2 Cn( An_[1] * C_[0] );
  matrix22 Amat;
  Amat =
    Cn(0), C_[Ncog-1](0),
    Cn(1), C_[Ncog-1](1);
  matrix22 Amatinv(inverse22(Amat));
  
  ainv_ = dzeromatrix(Ncog, 2);
  ainv_(0, 0) = Amatinv(0, 0);
  ainv_(0, 1) = Amatinv(0, 1);
  ainv_(Ncog-1, 0) = Amatinv(1, 0);
  ainv_(Ncog-1, 1) = Amatinv(1, 1);
}

void AnalyticalSolution::setBoundaryCondition(double k_[][3], dmatrix& ainv_, matrix22 *A_, matrix24 *B_, vector2 *C_, matrix22 *An_, matrix24 *Bn_, double *tm_, vector3 *zmp_des_, vector3 *vzmp_des_, vector3& xgi_cmd_, vector3& vgi_cmd_)
{
  for( int j = X ; j <= Y ; j++ ){
    vector2 Bcsum(0.0);
    vector4 Pzmp[Ncog];
    for( int i = 0 ; i < Ncog ; i++ ){
      Pzmp[i] = zmp_des_[i](j), vzmp_des_[i](j), zmp_des_[i+1](j), vzmp_des_[i+1](j);
      Bcsum += Bn_[i] * Pzmp[i];
    }
    
    vector2 Xgi(xgi_cmd_(j), vgi_cmd_(j));               // initial COG condition
    vector2 Xgf(zmp_des_[Ncog](j), vzmp_des_[Ncog](j));  // terminal COG condition
    vector2 Bc(Xgf - An_[0]*Xgi - Bcsum);
    dvector bvec(2);
    bvec(0) = Bc(0);
    bvec(1) = Bc(1);
    dvector P4(ublas::prod(ainv_, bvec));
    
    for( int i = 0 ; i < Ncog ; i++){
      double tm  = tm_[i+1];
      double tm2 = tm * tm;
      double tm3 = tm * tm2;
      
      matrix24 B13;
      B13 =
	-3.0/tm2, -2.0/tm,   3.0/tm2, -1.0/tm,
	 2.0/tm3,  1.0/tm2, -2.0/tm3,  1.0/tm2;
      
      vector2 C1;
      C1 =
         tm2,
	-2.0*tm;
      
      vector2 P23( B13 * Pzmp[i] + C1 * P4(i) );
      
      k_[i*7  ][j] = Xgi(0) - zmp_des_[i](j) - 2.0*P23(0)/omega2[i] - 24.0*P4(i)/omega4[i];
      k_[i*7+1][j] = (Xgi(1) - vzmp_des_[i](j) - 6.0*P23(1)/omega2[i])/omega[i];
      k_[i*7+2][j] = zmp_des_[i](j);
      k_[i*7+3][j] = vzmp_des_[i](j);
      k_[i*7+4][j] = P23(0);
      k_[i*7+5][j] = P23(1);
      k_[i*7+6][j] = P4(i);
      
      alias(Xgi) = A_[i] * Xgi + B_[i] * Pzmp[i] + C_[i] * P4(i);
    }
  }
  
  for( int i = 0 ; i < Ncog ; i++ ){
    double tm  = tm_[i+1];
    double tm2 = tm * tm;
    double tm3 = tm * tm2;
    double tm4 = tm * tm3;
    double tm5 = tm * tm4;

    const double ai = 0.0, af = 0.0;
    k_[i*7  ][Z] = zmp_des_[i](Z);
    k_[i*7+1][Z] = vzmp_des_[i](Z);
    k_[i*7+2][Z] = ai / 2.0;
    k_[i*7+3][Z] = ( 20.0*(zmp_des_[i+1](Z)-zmp_des_[i](Z)) - ( 8.0*vzmp_des_[i+1](Z)+12.0*vzmp_des_[i](Z))*tm + (    af-3.0*ai)*tm2) / (2.0*tm3);
    k_[i*7+4][Z] = (-30.0*(zmp_des_[i+1](Z)-zmp_des_[i](Z)) + (14.0*vzmp_des_[i+1](Z)+16.0*vzmp_des_[i](Z))*tm - (2.0*af-3.0*ai)*tm2) / (2.0*tm4);
    k_[i*7+5][Z] = ( 12.0*(zmp_des_[i+1](Z)-zmp_des_[i](Z)) -  6.0*(vzmp_des_[i+1](Z)+vzmp_des_[i](Z))*tm + (af-ai)*tm2) / (2.0*tm5);
    k_[i*7+6][Z] = 0.0;
  }
}

/////////////////////////////////////////////////////////////////////////////
// Adaptive Foot placement

void AnalyticalSolution::setCoefMatrix2(dmatrix &M_, matrix22 *A_, matrix24 *B_, vector2 *C_, matrix22 *An_, matrix24 *Bn_, double *tm_, WalkingState ws_)
{
  for( int i = 0 ; i < Ncog ; i++ ){
    const double tm  = tm_[i+1];
    const double tm2 = tm * tm;
    const double tm3 = tm * tm2;
    const double tm4 = tm * tm3;
    
    const double w  = omega[i];
    const double w2 = omega2[i];
    const double w4 = omega4[i];
    const double e = exp(w * tm);
    const double ch = (e * e + 1.0) / (2.0 * e);
    const double sh = (e * e - 1.0) / (2.0 * e);
    
    A_[i] =
      ch,   sh/w,
      w*sh, ch;
    
    matrix24 B11;
    B11 =
      1.0-ch, tm-sh/w, 0.0, 0.0,
      -w*sh,  1.0-ch,  0.0, 0.0;
    matrix22 B12;
    B12 =
      2.0*(1.0-ch)/w2+tm2, 6.0*(tm-sh/w)/w2 + tm3,
      2.0*(tm-sh/w),       6.0*(1.0-ch)/w2 + 3.0*tm2;
    matrix24 B13;
    B13 =
      -3.0/tm2, -2.0/tm, 3.0/tm2,  -1.0/tm,
      2.0/tm3,  1.0/tm2, -2.0/tm3, 1.0/tm2;
    B_[i] = B11 + B12 * B13;

    vector2 C1;
    C1 =
      tm2,
      -2.0*tm;
    vector2 C2;
    C2 =
      24.0*(1.0-ch)/w4 + 12.0*tm2/w2 + tm4,
      24.0*(tm-sh/w)/w2 + 4.0*tm3;
    C_[i] = B12 * C1 + C2;
  }
  
  An_[Ncog-1] = A_[Ncog-1];
  for( int i = Ncog-2 ; i >= 0 ; i-- )
    An_[i] = An_[i+1] * A_[i];
  
  for( int i = 0 ; i < Ncog-1 ; i++ )
    Bn_[i] = An_[i+1] * B_[i];
  Bn_[Ncog-1] = B_[Ncog-1];
  
  vector2 Bp[Ncog+1], Bv[Ncog+1];
  for( int i = 0 ; i < Ncog+1 ; i++ ){
    Bp[i] = 0.0;
    Bv[i] = 0.0;
  }
  for( int i = 0 ; i < Ncog ; i++ ){
    Bp[i] +=
      Bn_[i](0, 0),
      Bn_[i](1, 0);
    
    Bp[i+1] +=
      Bn_[i](0, 2),
      Bn_[i](1, 2);
    
    Bv[i] +=
      Bn_[i](0, 1),
      Bn_[i](1, 1);
    
    Bv[i+1] +=
      Bn_[i](0, 3),
      Bn_[i](1, 3);
  }
  
  int idx = 0;
  for( unsigned int i = 0 ; i < M_.size2() - 2 ; i++ ){
    M_(0, idx) = Bp[i*2+2](0) + Bp[i*2+3](0);
    M_(1, idx) = Bp[i*2+2](1) + Bp[i*2+3](1);
    idx++;
  }
  
  vector2 Cn( An_[1] * C_[0] );
  M_(0, idx) = Cn(0);
  M_(1, idx) = Cn(1);
  M_(0, idx+1) = C_[Ncog-1](0);
  M_(1, idx+1) = C_[Ncog-1](1);
}

void AnalyticalSolution::setBoundaryCondition2(double k_[][3], dmatrix& M_, dvector *s_mod, matrix22 *A_, matrix24 *B_, vector2 *C_, matrix22 *An_, matrix24 *Bn_, double *tm_, vector3 *zmp_des_, vector3 *vzmp_des_, vector3& xgi_cmd_, vector3& vgi_cmd_)
{
  dmatrix Q(didentity(M_.size2(), M_.size2()));
  Q(M_.size2()-2, M_.size2()-2) = 0.001;
  Q(M_.size2()-1, M_.size2()-1) = 0.001;
  
  dmatrix Qinv(Q.size1(), Q.size2());
  for( unsigned int i = 0 ; i < Q.size1(); i++ ){
    if( Q(i, i) != 0.0 )
      Qinv(i, i) = 1.0 / Q(i, i);
  }
  
  dmatrix MT(ublas::trans(M_));
  dmatrix QinvMT(ublas::prod(Qinv, MT));
  dmatrix MQinvMT(ublas::prod(M_, QinvMT));
  dmatrix MQinvMTinv(MQinvMT.size2(), MQinvMT.size1());
  calcPseudoInverse(MQinvMT, MQinvMTinv);
  

  const int Nstep_mod = M_.size2() - 2;
  for( int j = X ; j <= Y ; j++ ){
    
    // s_ref は絶対座標系
    dvector s_ref(dzerovector(M_.size2()));  
    for( int i = 0 ; i < Nstep_mod ; i++ )
      s_ref(i) = zmp_des_[2*i+2](j);
    
    vector2 Bcini(0.0);
    vector4 Pzmp[Ncog];
    
    for( int i = 0 ; i < Ncog ; i++ ){
      if( i == 0 )
	Pzmp[i] = zmp_des_[0](j), vzmp_des_[0](j), zmp_des_[1](j),   vzmp_des_[1](j);
      else if( i == 1 )
	Pzmp[i] = zmp_des_[1](j), vzmp_des_[1](j), 0.0,              vzmp_des_[2](j);
      else if( i > 1 && i < 2*Nstep_mod + 1 )
	Pzmp[i] = 0.0,            vzmp_des_[i](j), 0.0,              vzmp_des_[i+1](j);
      else if( i == 2*Nstep_mod + 1 )
	Pzmp[i] = 0.0,            vzmp_des_[i](j), zmp_des_[i+1](j), vzmp_des_[i+1](j);	
      else if( i > 2*Nstep_mod + 1 )
	Pzmp[i] = zmp_des_[i](j), vzmp_des_[i](j), zmp_des_[i+1](j), vzmp_des_[i+1](j);
      
      for( int i = 0 ; i < Ncog ; i++ )
	Bcini += Bn_[i] * Pzmp[i];
    }
    
    // Bv*v 常に0なので省略
    
    vector2 Xgi(xgi_cmd_(j), vgi_cmd_(j));               // initial COG condition
    vector2 Xgf(zmp_des_[Ncog](j), vzmp_des_[Ncog](j));  // terminal COG condition
    vector2 Bc(Xgf - An_[0]*Xgi - Bcini);
    dvector bvec(2);
    bvec(0) = Bc(0);
    bvec(1) = Bc(1);
    
#if 1
    dvector MsN(ublas::prod(M_, s_ref) - bvec);
    dvector MQinvMTinvMsN(ublas::prod(MQinvMTinv, MsN));
    s_mod[j] = -ublas::prod(QinvMT, MQinvMTinvMsN);
#else
    s_mod[j] = ublas::prod(ublas::prod(QinvMT, MQinvMTinv), bvec) - s_ref;
#endif
    
    dvector P4(Ncog);
    P4 = dzerovector(Ncog);
    
    P4(0)      = s_mod[j](Q.size2()-2);
    P4(Ncog-1) = s_mod[j](Q.size2()-1);
    
    for( int i = 0 ; i < Nstep_mod*2+1 ; i++ )
      zmp_des_[i+2](j) = s_ref(i/2) - s_mod[j](i/2);
    
    for( int i = 0 ; i < Ncog ; i++ )
      Pzmp[i] = zmp_des_[i](j), vzmp_des_[i](j), zmp_des_[i+1](j), vzmp_des_[i+1](j);
    
    for( int i = 0 ; i < Ncog ; i++){
      double tm  = tm_[i+1];
      double tm2 = tm * tm;
      double tm3 = tm * tm2;
      
      matrix24 B13;
      B13 =
	-3.0/tm2, -2.0/tm,   3.0/tm2, -1.0/tm,
	 2.0/tm3,  1.0/tm2, -2.0/tm3,  1.0/tm2;
      
      vector2 C1;
      C1 =
         tm2,
	-2.0*tm;
      
      vector2 P23( B13 * Pzmp[i] + C1 * P4(i) );
      
      k_[i*7  ][j] = Xgi(0) - zmp_des_[i](j) - 2.0*P23(0)/omega2[i] - 24.0*P4(i)/omega4[i];
      k_[i*7+1][j] = (Xgi(1) - vzmp_des_[i](j) - 6.0*P23(1)/omega2[i])/omega[i];
      k_[i*7+2][j] = zmp_des_[i](j);
      k_[i*7+3][j] = vzmp_des_[i](j);
      k_[i*7+4][j] = P23(0);
      k_[i*7+5][j] = P23(1);
      k_[i*7+6][j] = P4(i);
      
      alias(Xgi) = A_[i] * Xgi + B_[i] * Pzmp[i] + C_[i] * P4(i);
    }
  }
  
  for( int i = 0 ; i < Ncog ; i++ ){
    double tm  = tm_[i+1];
    double tm2 = tm * tm;
    double tm3 = tm * tm2;
    double tm4 = tm * tm3;
    double tm5 = tm * tm4;
    
    const double ai = 0.0, af = 0.0;
    k_[i*7  ][Z] = zmp_des_[i](Z);
    k_[i*7+1][Z] = vzmp_des_[i](Z);
    k_[i*7+2][Z] = ai / 2.0;
    k_[i*7+3][Z] = ( 20.0*(zmp_des_[i+1](Z)-zmp_des_[i](Z)) - ( 8.0*vzmp_des_[i+1](Z)+12.0*vzmp_des_[i](Z))*tm + (    af-3.0*ai)*tm2) / (2.0*tm3);
    k_[i*7+4][Z] = (-30.0*(zmp_des_[i+1](Z)-zmp_des_[i](Z)) + (14.0*vzmp_des_[i+1](Z)+16.0*vzmp_des_[i](Z))*tm - (2.0*af-3.0*ai)*tm2) / (2.0*tm4);
    k_[i*7+5][Z] = ( 12.0*(zmp_des_[i+1](Z)-zmp_des_[i](Z)) -  6.0*(vzmp_des_[i+1](Z)+vzmp_des_[i](Z))*tm + (af-ai)*tm2) / (2.0*tm5);
    k_[i*7+6][Z] = 0.0;
  }
}
