/*
 * Copyright (c) 2009,
 * @author Mitsuharu Morisawa
 *
 * AIST
 *
 * All rights reserved.
 *
 * This program is made available under the terms of the Eclipse Public License
 * v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 */
#ifndef __MCOG_ANALYTICAL_H__
#define __MCOG_ANALYTICAL_H__

#include "CogPatternGeneratorBase.h"


namespace mwalk_pattern_generator{
  
#define FAST_COG_TRAJECTORY
  
  class AnalyticalSolution : virtual public CogPatternGenerator_base{
    ANmatrix A;
    
  public:
    AnalyticalSolution(void);
    ~AnalyticalSolution();
    
    void makeZMPError(dvector *zmp_err_, vector3 *zmp_cmd_, double kg_[][3], double *tm_);
    void makeCoefParam(ANvector *bg_, double *tm_, vector3 *zmp_cmd_, vector3 *vzmp_cmd_, vector3& xgi_cmd_, vector3& vgi_cmd_);
    void makeCoefMatrix(ANmatrix& aginv_, double *tm_);
    void makeCoefVector(double kg_[][3], ANmatrix& aginv_, ANvector *bg_, double *tm_, vector3 *zmp_cmd_, vector3 *vzmp_cmd_);
    
    void setCoefMatrix(dmatrix &ainv_, matrix22 *A_, matrix24 *B_, vector2 *C_, matrix22 *An_, matrix24 *Bn_, double *tm_, WalkingState ws_);
    void setBoundaryCondition(double k_[][3], dmatrix& ainv_,
			      matrix22 *A_, matrix24 *B_, vector2 *C_, matrix22 *An_, matrix24 *Bn_,
			      double *tm_, vector3 *zmp_des_, vector3 *vzmp_des_, vector3& xgi_cmd_, vector3& vgi_cmd_);
    
    void setCoefMatrix2(dmatrix &M_, matrix22 *A_, matrix24 *B_, vector2 *C_, matrix22 *An_, matrix24 *Bn_, double *tm_, WalkingState ws_);
    void setBoundaryCondition2(double k_[][3], dmatrix& M_, dvector *s_mod,
			       matrix22 *A_, matrix24 *B_, vector2 *C_, matrix22 *An_, matrix24 *Bn_,
			       double *tm_, vector3 *zmp_des_, vector3 *vzmp_des_, vector3& xgi_cmd_, vector3& vgi_cmd_);
    
    void getMotion(double kg_[][3], vector3& x_cmd, vector3& v_cmd, vector3& a_cmd, vector3& zmp_cmd, vector3& vzmp_cmd, double *tm_, double tnow, int n);
    void getMotion(double kg_[][3], vector3& x_cmd, vector3& v_cmd, vector3& zmp_cmd, vector3& vzmp_cmd, double *tm_, double tnow, int n);
  };
  
} // end of namespace mwalk_pattern_generator

#endif // __MCOG_ANALYTICAL_H__
