/*
 * Copyright (c) 2009,
 * @author Mitsuharu Morisawa
 *
 * AIST
 *
 * All rights reserved.
 *
 * This program is made available under the terms of the Eclipse Public License
 * v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 */
#include "CogPatternGeneratorBase.h"

using namespace mwalk_pattern_generator;


CogPatternGenerator_base::CogPatternGenerator_base(void)
{
  for( int i = 0 ; i < Ncog ; i++ ){
    omega2[i] = Gv / (-INITIAL_ZMP_REF_Z);
    omega[i]  = sqrt(omega2[i]);
    omega4[i] = omega2[i]*omega2[i];
  }
  
  Foot_ini[LLEG](Y) =  DEFAULT_FOOT_DISTANCE / 2.0;
  Foot_ini[RLEG](Y) = -DEFAULT_FOOT_DISTANCE / 2.0;

  Foot[LLEG](Y) =  DEFAULT_FOOT_DISTANCE / 2.0;
  Foot[RLEG](Y) = -DEFAULT_FOOT_DISTANCE / 2.0;
  
#ifdef WITH_TOE_JOINT
  Zmp_toe[LLEG] = zmp_toe(X),  zmp_toe(Y), zmp_toe(Z);
  Zmp_toe[RLEG] = zmp_toe(X), -zmp_toe(Y), zmp_toe(Z);
#endif
  Zmp_ankle_ini[LLEG] = zmp_ankle(X),  zmp_ankle(Y), zmp_ankle(Z);
  Zmp_ankle_ini[RLEG] = zmp_ankle(X), -zmp_ankle(Y), zmp_ankle(Z);
  
  Zmp_ankle_end[LLEG] = zmp_ankle(X),  zmp_ankle(Y), zmp_ankle(Z);
  Zmp_ankle_end[RLEG] = zmp_ankle(X), -zmp_ankle(Y), zmp_ankle(Z);
  
  Zmp_last = 0.0;
}

CogPatternGenerator_base::~CogPatternGenerator_base(void)
{
}

void CogPatternGenerator_base::setCOGHeight(double *cog_height_)
{
  for( int i = 0 ; i < Ncog ; i++ ){
    if( cog_height_[i] <= 0.0 ){
      cerr << "Invalid cog_height[" << i << "]" << endl;
      cog_height_[i] = DEFAULT_COG_HEIGHT;
    }
    omega2[i] = Gv / cog_height_[i];
    omega[i]  = sqrt(omega2[i]);
    omega4[i] = omega2[i]*omega2[i];
  }
}

void CogPatternGenerator_base::setInitialFootDistance(vector3& foot_distance_ini)
{
  Foot_ini[LLEG] =  foot_distance_ini / 2.0;
  Foot_ini[RLEG] = -foot_distance_ini / 2.0;
}

void CogPatternGenerator_base::setFootDistance(double foot_distance)
{
  Foot[LLEG] = 0.0,  foot_distance / 2.0, 0.0;
  Foot[RLEG] = 0.0, -foot_distance / 2.0, 0.0;
}

void CogPatternGenerator_base::setZMPankle(vector3& zmp_ankle_)
{
  Zmp_ankle_ini[LLEG] = zmp_ankle_(X),  zmp_ankle_(Y), zmp_ankle_(Z);
  Zmp_ankle_ini[RLEG] = zmp_ankle_(X), -zmp_ankle_(Y), zmp_ankle_(Z);

  Zmp_ankle_end[LLEG] = zmp_ankle_(X),  zmp_ankle_(Y), zmp_ankle_(Z);
  Zmp_ankle_end[RLEG] = zmp_ankle_(X), -zmp_ankle_(Y), zmp_ankle_(Z);
}

void CogPatternGenerator_base::setZMPankles(vector3 *zmp_ankle_)
{
  Zmp_ankle_ini[LLEG] = zmp_ankle_[LLEG];
  Zmp_ankle_ini[RLEG] = zmp_ankle_[RLEG];

  Zmp_ankle_end[LLEG] = zmp_ankle_[LLEG];
  Zmp_ankle_end[RLEG] = zmp_ankle_[RLEG];
}

void CogPatternGenerator_base::setZMPankle(vector3& zmp_ankle_ini_, vector3& zmp_ankle_end_)
{
  Zmp_ankle_ini[LLEG] = zmp_ankle_ini_(X),  zmp_ankle_ini_(Y), zmp_ankle_ini_(Z);
  Zmp_ankle_ini[RLEG] = zmp_ankle_ini_(X), -zmp_ankle_ini_(Y), zmp_ankle_ini_(Z);

  Zmp_ankle_end[LLEG] = zmp_ankle_end_(X),  zmp_ankle_end_(Y), zmp_ankle_end_(Z);
  Zmp_ankle_end[RLEG] = zmp_ankle_end_(X), -zmp_ankle_end_(Y), zmp_ankle_end_(Z);
}

void CogPatternGenerator_base::setZMPankles(vector3 *zmp_ankle_ini_, vector3 *zmp_ankle_end_)
{
  Zmp_ankle_ini[LLEG] = zmp_ankle_ini_[LLEG];
  Zmp_ankle_ini[RLEG] = zmp_ankle_ini_[RLEG];
  
  Zmp_ankle_end[LLEG] = zmp_ankle_end_[LLEG];
  Zmp_ankle_end[RLEG] = zmp_ankle_end_[RLEG];
}

#ifdef WITH_TOE_JOINT
void CogPatternGenerator_base::setZMPtoe(vector3& zmp_toe_)
{
  Zmp_toe[LLEG] = zmp_toe_(X),  zmp_toe_(Y), zmp_toe_(Z);
  Zmp_toe[RLEG] = zmp_toe_(X), -zmp_toe_(Y), zmp_toe_(Z);
}

void CogPatternGenerator_base::setZMPtoes(vector3 *zmp_toe_)
{
  Zmp_toe[LLEG] = zmp_toe_[LLEG];
  Zmp_toe[RLEG] = zmp_toe_[RLEG];
}
#endif

#ifndef WITH_TOE_JOINT

void CogPatternGenerator_base::predictZMPPath(double *Tm_, vector3 *Zmp_, vector3 *VZmp_, vector3 Zmpi_, vector3 VZmpi_, vector3& AbsPosOrg_, vector3& AbsRpyOrg_, deque<FootPlace>& FootQueue_, WalkingState ws_, int n_period_, int n_now_)
{
  FootQueue_.front();
  unsigned int FQsize = FootQueue_.size();
  if( FQsize > Nstep ) FQsize = Nstep;
  
  unsigned int foot_step = 0;
  
  double   total_angle = AbsRpyOrg_(Z);
  matrix33 Rot( rotFromRpy(0.0, 0.0, total_angle) );
  vector3  total_step( AbsPosOrg_ );
  vector3  zmp_end;

  Tm_[0] = (double)n_now_*DT;
  Zmp_[0] = Zmpi_;
  VZmp_[0] = VZmpi_;
  
  for( int i = 1 ; i <= Ncog ; i++ ){
    Tm_[i] = (double)n_period_*DT;
    
    switch( ws_ ){
    case STAND_ANKLE:
      Zmpi_ = total_step + Rot * ((Zmp_ankle_ini[RLEG]+Zmp_ankle_ini[LLEG])/2.0);
      if( foot_step >=  FQsize ){
	n_period_ = Nss;
	ws_ = STAND_ANKLE;
      }
      else{
	n_period_ = FootQueue_[foot_step].nds;	
	if( FootQueue_[foot_step].support_leg != RLEG ){
	  ws_ = CENTER_TO_LEFT_ANKLE;
	  total_step += Rot * (Foot_ini[LLEG] - Foot[LLEG]);
	}
	else{
	  ws_ = CENTER_TO_RIGHT_ANKLE;
	  total_step += Rot * (Foot_ini[RLEG] - Foot[RLEG]);
	}
      }
      break;
      
    case CENTER_TO_LEFT_ANKLE:
    case RIGHT_TO_LEFT_ANKLE:
      Zmpi_ = total_step + Rot * (Foot[LLEG] + rotFromRpy(FootQueue_[foot_step].rpy) * Zmp_ankle_ini[LLEG]);
      n_period_ = FootQueue_[foot_step].nss;
      ws_ = LEFT_SUPPORT_ANKLE;
      break;
      
    case KEEP_LEFT_ANKLE:
      n_period_ = FootQueue_[foot_step].nss;
      ws_ = LEFT_SUPPORT_ANKLE;
      break;
      
    case CENTER_TO_RIGHT_ANKLE:
    case LEFT_TO_RIGHT_ANKLE:
      Zmpi_ = total_step + Rot * (Foot[RLEG] + rotFromRpy(FootQueue_[foot_step].rpy) * Zmp_ankle_ini[RLEG]);
      n_period_ = FootQueue_[foot_step].nss;
      ws_ = RIGHT_SUPPORT_ANKLE;
      break;
      
    case KEEP_RIGHT_ANKLE:
      n_period_ = FootQueue_[foot_step].nss;
      ws_ = RIGHT_SUPPORT_ANKLE;      
      break;
      
    case LEFT_TO_CENTER_ANKLE:
    case RIGHT_TO_CENTER_ANKLE:
      Zmpi_ = total_step + Rot * ((Zmp_ankle_ini[RLEG]+Zmp_ankle_ini[LLEG])/2.0 + Zmp_last);
      n_period_ = Nss;
      ws_ = STAND_ANKLE;
      break;
      
    case LEFT_SUPPORT_ANKLE:
      zmp_end = total_step + Rot * (Foot[LLEG] + rotFromRpy(FootQueue_[foot_step].rpy) * Zmp_ankle_end[LLEG]);
#ifdef ORIGIN_PATH
      if( FQsize == 1 )
	total_step += Rot * FootQueue_[foot_step].p / 2.0;
      else
	total_step += Rot * FootQueue_[foot_step].p;
#else
      if( FQsize == 1 )
	total_step += Rot * FootQueue_[foot_step].p / 2.0;
      else
	total_step += Rot * (FootQueue_[foot_step].p + Foot[LLEG] - rotFromRpy(FootQueue_[foot_step].rpy) * Foot[RLEG]);
#endif
      total_angle += FootQueue_[foot_step].rpy(Z);
      Rot = rotFromRpy(0.0, 0.0, total_angle);      
      foot_step++;
      if( foot_step < FQsize ){
	n_period_ = FootQueue_[foot_step].nds;
	if( FootQueue_[foot_step].support_leg != RLEG )
	  ws_ = KEEP_LEFT_ANKLE;
	else{
	  ws_ = LEFT_TO_RIGHT_ANKLE;
	  Zmpi_ = zmp_end;
	}
      }
      else{
	n_period_ = Nds;
	ws_ = LEFT_TO_CENTER_ANKLE;
      }
      break;
      
    case RIGHT_SUPPORT_ANKLE:
      zmp_end = total_step + Rot * (Foot[RLEG] + rotFromRpy(FootQueue_[foot_step].rpy) * Zmp_ankle_end[RLEG]);
#ifdef ORIGIN_PATH
      if( FQsize == 1 )
	total_step += Rot * FootQueue_[foot_step].p / 2.0;
      else
	total_step += Rot * FootQueue_[foot_step].p;
#else
      if( FQsize == 1 )
	total_step += Rot * FootQueue_[foot_step].p / 2.0;
      else
	total_step += Rot * (FootQueue_[foot_step].p + Foot[RLEG] - rotFromRpy(FootQueue_[foot_step].rpy) * Foot[LLEG]);
#endif
      total_angle += FootQueue_[foot_step].rpy(Z);
      Rot = rotFromRpy(0.0, 0.0, total_angle);
      foot_step++;
      if( foot_step < FQsize ){
	n_period_ = FootQueue_[foot_step].nds;
	if( FootQueue_[foot_step].support_leg != RLEG ){
	  ws_ = RIGHT_TO_LEFT_ANKLE;
	  Zmpi_ = zmp_end;
	}
	else
	  ws_ = KEEP_RIGHT_ANKLE;
      }
      else{
	n_period_ = Nds;
	ws_ = RIGHT_TO_CENTER_ANKLE;
      }
      break;
      
    default:
      cerr << "Warning!!! Invalid walking state!!!" << endl;
      break;
    }

    Zmp_[i] = Zmpi_;
  }

#if 0
  for( int i = 0 ; i <= Ncog ; i++ )
    std::cout << "Tm[" << i << "]=" << Tm_[i] << ", Zmp_pas[" << i << "]=" << Zmp_[i] << endl;
#endif
}

#else // end of '#ifndef WITH_TOE_JOINT'

void CogPatternGenerator_base::predictZMPPath(double *Tm_, vector3 *Zmp_, vector3 *VZmp_, vector3 Zmpi_, vector3 VZmpi_, vector3& AbsPosOrg_, vector3& AbsRpyOrg_, deque<FootPlace>& FootQueue_, WalkingState ws_, WalkingState ws_last_, int n_period_, int n_now_)
{
  FootQueue_.front();
  unsigned int FQsize = FootQueue_.size();
  if( FQsize > Nstep ) FQsize = Nstep;

  unsigned int foot_step = 0;

  double   total_angle = AbsRpyOrg_(Z);
  matrix33 Rot( rotFromRpy(0.0, 0.0, total_angle) );
  vector3  total_step( AbsPosOrg_ );

  Tm_[0] = (double)n_now_*DT;
  Zmp_[0] = Zmpi_;
  VZmp_[0] = VZmpi_;
  
  for( int i = 1 ; i <= Ncog ; i++ ){
    Tm_[i] = (double)n_period_*DT;

    switch( ws_ ){
    case STAND_ANKLE:
      Zmpi_ = total_step + Rot * ((Zmp_ankle_ini[RLEG]+Zmp_ankle_ini[LLEG])/2.0);
      ws_last_ = ws_;
      if( foot_step >=  FQsize ){
	n_period_ = Nds;
	ws_ = STAND_ANKLE;
      }
      else{
	n_period_ = FootQueue_[foot_step].nds;	
	if( FootQueue_[foot_step].support_leg != RLEG )
	  ws_ = CENTER_TO_LEFT_ANKLE;
	else
	  ws_ = CENTER_TO_RIGHT_ANKLE;
      }
      break;
      
    case CENTER_TO_LEFT_ANKLE:
    case RIGHT_TO_LEFT_ANKLE:
      ws_last_ = ws_;
      if( FootQueue_[foot_step].p(X) >= 0.0 ){
	n_period_ = NPzmp_ankle_to_toe(FootQueue_[foot_step].nss, 0);
	Zmpi_ = total_step + Rot * (Foot[LLEG] + rotFromRpy(FootQueue_[foot_step].rpy) * Zmp_ankle_ini[LLEG]);
	ws_ = LEFT_SUPPORT_ANKLE;
      }
      else{
	n_period_ = NPzmp_toe_to_ankle(FootQueue_[foot_step].nss, 0);
	Zmpi_ = total_step + Rot * (Foot[LLEG] + rotFromRpy(FootQueue_[foot_step].rpy) * Zmp_toe[LLEG]);
	ws_ = LEFT_SUPPORT_TOE;
      }
      break;
      
    case KEEP_LEFT_ANKLE:
      ws_last_ = ws_;
      if( FootQueue_[foot_step].p(X) >= 0.0 ){
	n_period_ = NPzmp_ankle_to_toe(FootQueue_[foot_step].nss, 0);
	ws_ = LEFT_SUPPORT_ANKLE;
      }
      else{
	n_period_ = NPzmp_toe_to_ankle(FootQueue_[foot_step].nss, 0);
	ws_ = LEFT_SUPPORT_TOE;
      }
      break;
      
    case CENTER_TO_RIGHT_ANKLE:
    case LEFT_TO_RIGHT_ANKLE:
      ws_last_ = ws_;
      if( FootQueue_[foot_step].p(X) >= 0.0 ){
	n_period_ = NPzmp_ankle_to_toe(FootQueue_[foot_step].nss, 0);	
	Zmpi_ = total_step + Rot * (Foot[RLEG] + rotFromRpy(FootQueue_[foot_step].rpy) * Zmp_ankle_ini[RLEG]);
	ws_ = RIGHT_SUPPORT_ANKLE;
      }
      else{
	n_period_ = NPzmp_toe_to_ankle(FootQueue_[foot_step].nss, 0);	
	Zmpi_ = total_step + Rot * (Foot[RLEG] + rotFromRpy(FootQueue_[foot_step].rpy) * Zmp_toe[RLEG]);
	ws_ = RIGHT_SUPPORT_TOE;
      }
      break;
      
    case KEEP_RIGHT_ANKLE:
      ws_last_ = ws_;
      if( FootQueue_[foot_step].p(X) >= 0.0 ){
	n_period_ = NPzmp_ankle_to_toe(FootQueue_[foot_step].nss, 0);	
	ws_ = RIGHT_SUPPORT_ANKLE;
      }
      else{
	n_period_ = NPzmp_toe_to_ankle(FootQueue_[foot_step].nss, 0);	
	ws_ = RIGHT_SUPPORT_TOE;
      }
      break;
      
    case LEFT_TO_CENTER_ANKLE:
    case RIGHT_TO_CENTER_ANKLE:
      Zmpi_ = total_step + Rot * ((Zmp_ankle_ini[RLEG]+Zmp_ankle_ini[LLEG])/2.0 + Zmp_last);     
      n_period_ = Nss;
      ws_last_ = ws_;
      ws_ = STAND_ANKLE;
      break;


    case LEFT_SUPPORT_ANKLE:
      if( ws_last_ == LEFT_SUPPORT_TOE_TO_ANKLE || ws_last_ == KEEP_LEFT_TOE_TO_ANKLE ){
#ifdef ORIGIN_PATH
	total_step += Rot * FootQueue_[foot_step].p;
#else
	total_step += Rot * (FootQueue_[foot_step].p + Foot[LLEG] - rotFromRpy(FootQueue_[foot_step].rpy) * Foot[RLEG]);
#endif
	total_angle += FootQueue_[foot_step].rpy(Z);
	Rot = rotFromRpy(0.0, 0.0, total_angle);      
	foot_step++;
	ws_last_ = ws_;
	if( foot_step < FQsize ){
	  n_period_ = FootQueue_[foot_step].nds;
	  if( FootQueue_[foot_step].support_leg != RLEG )
	    ws_ = KEEP_LEFT_ANKLE;
	  else
	    ws_ = LEFT_TO_RIGHT_ANKLE;
	}
	else{
	  n_period_ = Nds;
	  ws_ = LEFT_TO_CENTER_ANKLE;
	}
      }
      else{
	n_period_ = NPzmp_ankle_to_toe(FootQueue_[foot_step].nss, 1) - NPzmp_ankle_to_toe(FootQueue_[foot_step].nss, 0);
	if( ws_last_ == KEEP_LEFT_ANKLE ){
	  ws_last_ = ws_;
	  ws_ = KEEP_LEFT_ANKLE_TO_TOE;
	}
	else{
	  ws_last_ = ws_;
	  ws_ = LEFT_SUPPORT_ANKLE_TO_TOE;
	}
      }
      break;
      
    case LEFT_SUPPORT_TOE:
      if( ws_last_ == LEFT_SUPPORT_ANKLE_TO_TOE || ws_last_ == KEEP_LEFT_ANKLE_TO_TOE ){
#ifdef ORIGIN_PATH
	total_step += Rot * FootQueue_[foot_step].p;
#else
	total_step += Rot * (FootQueue_[foot_step].p + Foot[LLEG] - rotFromRpy(FootQueue_[foot_step].rpy) * Foot[RLEG]);
#endif
	total_angle += FootQueue_[foot_step].rpy(Z);
	Rot = rotFromRpy(0.0, 0.0, total_angle);      
	foot_step++;
	ws_last_ = ws_;
	if( foot_step < FQsize ){
	  n_period_ = FootQueue_[foot_step].nds;
	  if( FootQueue_[foot_step].support_leg != RLEG )
	    ws_ = KEEP_LEFT_ANKLE;
	  else
	    ws_ = LEFT_TO_RIGHT_ANKLE;
	}
	else{
	  n_period_ = Nds;
	  ws_ = LEFT_TO_CENTER_ANKLE;
	}
      }
      else{
	n_period_ = NPzmp_toe_to_ankle(FootQueue_[foot_step].nss, 1) - NPzmp_toe_to_ankle(FootQueue_[foot_step].nss, 0);
	if( ws_last_ == KEEP_LEFT_ANKLE || ws_last_ == KEEP_LEFT_TOE ){
	  ws_last_ = ws_;
	  ws_ = KEEP_LEFT_ANKLE_TO_TOE;
	}
	else{
	  ws_last_ = ws_;
	  ws_ = LEFT_SUPPORT_TOE_TO_ANKLE;
	}
      }
      break;
      
    case LEFT_SUPPORT_ANKLE_TO_TOE:
      n_period_ = FootQueue_[foot_step].nss - NPzmp_ankle_to_toe(FootQueue_[foot_step].nss, 1);
      if( foot_step < FQsize - 1 )
	Zmpi_ = total_step + Rot * (Foot[LLEG] + rotFromRpy(FootQueue_[foot_step].rpy) * Zmp_toe[LLEG]);
      else
	Zmpi_ = total_step + Rot * (Foot[LLEG] + rotFromRpy(FootQueue_[foot_step].rpy) * Zmp_ankle_ini[LLEG]);
      ws_last_ = ws_;
      ws_ = LEFT_SUPPORT_TOE;
      break;      
      
    case LEFT_SUPPORT_TOE_TO_ANKLE:
      n_period_ = FootQueue_[foot_step].nss - NPzmp_toe_to_ankle(FootQueue_[foot_step].nss, 1);
      Zmpi_ = total_step + Rot * (Foot[LLEG] + rotFromRpy(FootQueue_[foot_step].rpy) * Zmp_ankle_ini[LLEG]);
      ws_last_ = ws_;
      ws_ = LEFT_SUPPORT_ANKLE;      
      break;
      
    case KEEP_LEFT_ANKLE_TO_TOE:
      n_period_ = FootQueue_[foot_step].nss - NPzmp_ankle_to_toe(FootQueue_[foot_step].nss, 1);
      ws_last_ = ws_;
      ws_ = LEFT_SUPPORT_TOE;
      break;
      
    case KEEP_LEFT_TOE_TO_ANKLE:
      n_period_ = FootQueue_[foot_step].nss - NPzmp_toe_to_ankle(FootQueue_[foot_step].nss, 1);
      ws_last_ = ws_;
      ws_ = LEFT_SUPPORT_ANKLE;      
      break;
      
    case RIGHT_SUPPORT_ANKLE:
      if( ws_last_ == RIGHT_SUPPORT_TOE_TO_ANKLE || ws_last_ == KEEP_RIGHT_TOE_TO_ANKLE ){
#ifdef ORIGIN_PATH
	total_step += Rot * FootQueue_[foot_step].p;
#else
	total_step += Rot * (FootQueue_[foot_step].p + Foot[RLEG] - rotFromRpy(FootQueue_[foot_step].rpy) * Foot[LLEG]);
#endif
	total_angle += FootQueue_[foot_step].rpy(Z);
	Rot = rotFromRpy(0.0, 0.0, total_angle);      
	foot_step++;
	ws_last_ = ws_;
	if( foot_step < FQsize ){
	  n_period_ = FootQueue_[foot_step].nds;
	  if( FootQueue_[foot_step].support_leg != RLEG )
	    ws_ = RIGHT_TO_LEFT_ANKLE;
	  else
	    ws_ = KEEP_RIGHT_ANKLE;
	}
	else{
	  n_period_ = Nds;
	  ws_ = RIGHT_TO_CENTER_ANKLE;
	}
      }
      else{
	n_period_ = NPzmp_ankle_to_toe(FootQueue_[foot_step].nss, 1) - NPzmp_ankle_to_toe(FootQueue_[foot_step].nss, 0);
	if( ws_last_ == KEEP_RIGHT_ANKLE || ws_last_ == KEEP_RIGHT_TOE ){
	  ws_last_ = ws_;
	  ws_ = KEEP_RIGHT_ANKLE_TO_TOE;
	}
	else{
	  ws_last_ = ws_;
	  ws_ = RIGHT_SUPPORT_ANKLE_TO_TOE;
	}
      }
      break;
      
    case RIGHT_SUPPORT_TOE:
      if( ws_last_ == RIGHT_SUPPORT_ANKLE_TO_TOE || ws_last_ == KEEP_RIGHT_ANKLE_TO_TOE ){
#ifdef ORIGIN_PATH
	total_step += Rot * FootQueue_[foot_step].p;
#else
	total_step += Rot * (FootQueue_[foot_step].p + Foot[RLEG] - rotFromRpy(FootQueue_[foot_step].rpy) * Foot[LLEG]);
#endif
	total_angle += FootQueue_[foot_step].rpy(Z);
	Rot = rotFromRpy(0.0, 0.0, total_angle);      
	foot_step++;
	ws_last_ = ws_;
	if( foot_step < FQsize ){
	  n_period_ = FootQueue_[foot_step].nds;
	  if( FootQueue_[foot_step].support_leg != RLEG )
	    ws_ = RIGHT_TO_LEFT_ANKLE;
	  else
	    ws_ = KEEP_RIGHT_ANKLE;
	}
	else{
	  n_period_ = Nds;
	  ws_ = RIGHT_TO_CENTER_ANKLE;
	}
      }
      else{
	n_period_ = NPzmp_toe_to_ankle(FootQueue_[foot_step].nss, 1) - NPzmp_toe_to_ankle(FootQueue_[foot_step].nss, 0);
	if( ws_last_ == KEEP_RIGHT_ANKLE || ws_last_ == KEEP_RIGHT_TOE ){
	  ws_last_ = ws_;
	  ws_ = KEEP_RIGHT_TOE_TO_ANKLE;
	}
	else{
	  ws_last_ = ws_;
	  ws_ = RIGHT_SUPPORT_TOE_TO_ANKLE;
	}
      }
      break;
      
    case RIGHT_SUPPORT_ANKLE_TO_TOE:
      n_period_ = FootQueue_[foot_step].nss - NPzmp_ankle_to_toe(FootQueue_[foot_step].nss, 1);
      if( foot_step < FQsize - 1 )
	Zmpi_ = total_step + Rot * (Foot[RLEG] + rotFromRpy(FootQueue_[foot_step].rpy) * Zmp_toe[RLEG]);
      else
	Zmpi_ = total_step + Rot * (Foot[RLEG] + rotFromRpy(FootQueue_[foot_step].rpy) * Zmp_ankle_ini[RLEG]);
      ws_last_ = ws_;
      ws_ = RIGHT_SUPPORT_TOE;
      break;      
      
    case RIGHT_SUPPORT_TOE_TO_ANKLE:
      n_period_ = FootQueue_[foot_step].nss - NPzmp_toe_to_ankle(FootQueue_[foot_step].nss, 1);
      Zmpi_ = total_step + Rot * (Foot[RLEG] + rotFromRpy(FootQueue_[foot_step].rpy) * Zmp_ankle_ini[RLEG]);
      ws_last_ = ws_;
      ws_ = RIGHT_SUPPORT_ANKLE;      
      break;
      
    case KEEP_RIGHT_ANKLE_TO_TOE:
      n_period_ = FootQueue_[foot_step].nss - NPzmp_ankle_to_toe(FootQueue_[foot_step].nss, 1);
      ws_last_ = ws_;
      ws_ = RIGHT_SUPPORT_TOE;
      break;      
      
    case KEEP_RIGHT_TOE_TO_ANKLE:
      n_period_ = FootQueue_[foot_step].nss - NPzmp_toe_to_ankle(FootQueue_[foot_step].nss, 1);
      ws_last_ = ws_;
      ws_ = RIGHT_SUPPORT_ANKLE;      
      break;
      
    default:
      cerr << "Warning!!! Invalid walking state!!!" << endl;
      break;
    }

    Zmp_[i] = Zmpi_;
  }

#if 0
  for( int i = 0 ; i <= Ncog ; i++ )
    std::cout << "Tm[" << i << "]=" << Tm_[i] << ", Zmp_pas[" << i << "]=" << Zmp_[i] << endl;
#endif
}

#endif // #ifndef WITH_TOE_JOINT #else
