/*
 * Copyright (c) 2009,
 * @author Mitsuharu Morisawa
 *
 * AIST
 *
 * All rights reserved.
 *
 * This program is made available under the terms of the Eclipse Public License
 * v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 */
#ifndef __MCOG_H__
#define __MCOG_H__

#include <queue>
#include "mwalk_common.h"
#include "FootPlace.h"


namespace mwalk_pattern_generator{
  
  class CogPatternGenerator_base{
  public:
    double  omega[Ncog];
    double  omega2[Ncog]; //omega^2 = g/z
    double  omega4[Ncog]; //omega^4
    vector3 Foot_ini[2];
    vector3 Foot[2];
    
    vector3 Zmp_ankle_ini[2];
    vector3 Zmp_ankle_end[2];
#ifdef WITH_TOE_JOINT
    vector3 Zmp_toe[2];
#endif
    vector3 Zmp_last;
    
    CogPatternGenerator_base(void);
    virtual ~CogPatternGenerator_base(void);
    
    void setCOGHeight(double *cog_height_);
    void setInitialFootDistance(vector3& foot_distance_ini);
    void setFootDistance(double foot_distance);
    void setZMPankle(vector3& zmp_ankle_);
    void setZMPankles(vector3 *zmp_ankle_);
    void setZMPankle(vector3& zmp_ankle_ini_, vector3& zmp_ankle_end_);
    void setZMPankles(vector3 *zmp_ankle_ini_, vector3 *zmp_ankle_end_);
#ifdef WITH_TOE_JOINT
    void setZMPtoe(vector3& zmp_toe_);
    void setZMPtoes(vector3 *zmp_toe_);
#endif
    void setZMPlast(vector3& zmp_last_){ Zmp_last = zmp_last_; };
#ifndef WITH_TOE_JOINT
    void predictZMPPath(double *Tm_, vector3 *Zmp_, vector3 *VZmp_, vector3 Zmpi_, vector3 VZmpi_, vector3& AbsPosOrg_, vector3& AbsRpyOrg_, deque<FootPlace>& FootQueue_, WalkingState ws_, int n_period_, int n_now_);
#else
    void predictZMPPath(double *Tm_, vector3 *Zmp_, vector3 *VZmp_, vector3 Zmpi_, vector3 VZmpi_, vector3& AbsPosOrg_, vector3& AbsRpyOrg_, deque<FootPlace>& FootQueue_, WalkingState ws_, WalkingState ws_last_, int n_period_, int n_now_);
#endif

  };
  
} // end of namespace mwalk_pattern_generator

#endif // __MCOG_H__
