/*
 * Copyright (c) 2009,
 * @author Mitsuharu Morisawa
 *
 * AIST
 *
 * All rights reserved.
 *
 * This program is made available under the terms of the Eclipse Public License
 * v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 */
#include "ModelManager.h"
#include "FootMotionGenerator.h"

using namespace motion_generator;
using namespace mwalk_pattern_generator;


FootMotionGenerator::FootMotionGenerator(void)
{
  swg_height_threshold = 0.03;
  swg_rpy_threshold = 0.05;
  
  R_blandoff[FLAT] = 0.0;
  R_blandoff[DOWNSTAIRS] = 0.0;
  R_blandoff[UPSTAIRS] = 0.0;
  
  R_btouchdown[FLAT] = 0.0;
  R_btouchdown[DOWNSTAIRS] = 0.0;
  R_btouchdown[UPSTAIRS] = 0.0;
  
  R_landoff[FLAT] = 0.0;
  R_landoff[DOWNSTAIRS] = 0.0;
  R_landoff[UPSTAIRS] = 0.0;
  
  R_touchdown[FLAT] = 0.0;
  R_touchdown[DOWNSTAIRS] = 0.0;
  R_touchdown[UPSTAIRS] = 0.0;
  
  Tprofile_horizontal[FLAT][0] = 0.072;
  Tprofile_horizontal[FLAT][1] = 0.358;
  Tprofile_horizontal[FLAT][2] = 0.643;
  Tprofile_horizontal[FLAT][3] = 0.858;
  
  Tprofile_vertical[FLAT][0] = 0;
  Tprofile_vertical[FLAT][1] = 0.143;
  Tprofile_vertical[FLAT][2] = 0.358;
  Tprofile_vertical[FLAT][3] = 0.5;
  Tprofile_vertical[FLAT][4] = 0.5;
  Tprofile_vertical[FLAT][5] = 0.643;
  Tprofile_vertical[FLAT][6] = 0.858;
  Tprofile_vertical[FLAT][7] = 1.0;
  
  TP_pivot[FLAT][0] = 0.215;
  TP_pivot[FLAT][1] = 0.786;
  
  TR_ankle[FLAT][0] = 0.0;
  TR_ankle[FLAT][1] = 0.215;
  TR_ankle[FLAT][2] = 0.786;
  TR_ankle[FLAT][3] = 1.0;
  
  XPmiddle_ratio[FLAT] = 0.5;
  SWmiddle_point[FLAT] = 0.05;
  
  Tprofile_horizontal[DOWNSTAIRS][0] = 0.072;
  Tprofile_horizontal[DOWNSTAIRS][1] = 0.286;
  Tprofile_horizontal[DOWNSTAIRS][2] = 0.536;
  Tprofile_horizontal[DOWNSTAIRS][3] = 0.643;
  
  Tprofile_vertical[DOWNSTAIRS][0] = 0;
  Tprofile_vertical[DOWNSTAIRS][1] = 0.143;
  Tprofile_vertical[DOWNSTAIRS][2] = 0.358;
  Tprofile_vertical[DOWNSTAIRS][3] = 0.5;
  Tprofile_vertical[DOWNSTAIRS][4] = 0.5;
  Tprofile_vertical[DOWNSTAIRS][5] = 0.643;
  Tprofile_vertical[DOWNSTAIRS][6] = 0.858;
  Tprofile_vertical[DOWNSTAIRS][7] = 1.0;
  
  TP_pivot[DOWNSTAIRS][0] = 0.179;
  TP_pivot[DOWNSTAIRS][1] = 0.643;
  
  TR_ankle[DOWNSTAIRS][0] = 0.0;
  TR_ankle[DOWNSTAIRS][1] = 0.179;
  TR_ankle[DOWNSTAIRS][2] = 0.643;
  TR_ankle[DOWNSTAIRS][3] = 1.0;
  
  XPmiddle_ratio[DOWNSTAIRS] = 0.25;
  SWmiddle_point[DOWNSTAIRS] = 0.05;
  
  Tprofile_horizontal[UPSTAIRS][0] = 0.215;
  Tprofile_horizontal[UPSTAIRS][1] = 0.465;
  Tprofile_horizontal[UPSTAIRS][2] = 0.643;
  Tprofile_horizontal[UPSTAIRS][3] = 0.858;
  
  Tprofile_vertical[UPSTAIRS][0] = 0;
  Tprofile_vertical[UPSTAIRS][1] = 0.143;
  Tprofile_vertical[UPSTAIRS][2] = 0.358;
  Tprofile_vertical[UPSTAIRS][3] = 0.5;
  Tprofile_vertical[UPSTAIRS][4] = 0.5;
  Tprofile_vertical[UPSTAIRS][5] = 0.643;
  Tprofile_vertical[UPSTAIRS][6] = 0.858;
  Tprofile_vertical[UPSTAIRS][7] = 1.0;
  
  TP_pivot[UPSTAIRS][0] = 0.358;
  TP_pivot[UPSTAIRS][1] = 0.822;
  
  TR_ankle[UPSTAIRS][0] = 0.0;
  TR_ankle[UPSTAIRS][1] = 0.358;
  TR_ankle[UPSTAIRS][2] = 0.822;
  TR_ankle[UPSTAIRS][3] = 1.0;
  
  XPmiddle_ratio[UPSTAIRS] = 1.0;
  SWmiddle_point[UPSTAIRS] = 0.03;
  
  for( int i = 0 ; i < 3 ; i++ ){
    SwingLegPos[i].InterpolatorType(VELOCITY_PROFILE);
    AnkleRpy[RLEG][i].InterpolatorType(POINT_TO_POINT);
    AnkleRpy[LLEG][i].InterpolatorType(POINT_TO_POINT);
  }
  
  for( int i = 0 ; i < 2 ; i++ ){
    Pivot[RLEG][i].InterpolatorType(POINT_TO_POINT);
    Pivot[LLEG][i].InterpolatorType(POINT_TO_POINT);
    SwingRpy[RLEG][i].InterpolatorType(CLAMPED_CUBIC_SPLINE, 4);
    SwingRpy[LLEG][i].InterpolatorType(CLAMPED_CUBIC_SPLINE, 4);
  }
  
#ifdef WITH_TOE_JOINT
  ToePitch[RLEG].InterpolatorType(POINT_TO_POINT);
  ToePitch[LLEG].InterpolatorType(POINT_TO_POINT);
#endif
}

FootMotionGenerator::~FootMotionGenerator()
{
}

void FootMotionGenerator::Reset(int n_clock, int n_period)
{
#ifdef WITH_TOE_JOINT
  ToePitch[RLEG].setMotionData(n_clock, n_clock+n_period, 0.0);
  ToePitch[LLEG].setMotionData(n_clock, n_clock+n_period, 0.0);
#endif
#if 0
  Pivot[RLEG][X].setMotionData(n_clock, n_clock+n_period, 0.0);
  Pivot[RLEG][Y].setMotionData(n_clock, n_clock+n_period, 0.0);
  Pivot[LLEG][X].setMotionData(n_clock, n_clock+n_period, 0.0);
  Pivot[LLEG][Y].setMotionData(n_clock, n_clock+n_period, 0.0);
#endif
}

int FootMotionGenerator::selectFloor(double step_height)
{
  if( step_height > swg_height_threshold )
    return UPSTAIRS;
  else if( step_height < -swg_height_threshold )
    return DOWNSTAIRS;
  return FLAT;
}

void FootMotionGenerator::setSwingPosParam(State& Leg_ini, State& Leg_end, int n_start, int n_period, int n_total)
{
  static int floor = FLAT;
  // Vertical motion
  if( n_period == n_total ){
    const double step_height = Leg_end.P(Z) - Leg_ini.P(Z);
    floor = selectFloor(step_height);
    
    SwingLegPos[Z].setMotionData(
      n_start + static_cast<int>(Tprofile_vertical[floor][0]*n_total),
      Leg_ini.P(Z),
      Leg_ini.V(Z),
      Leg_ini.Vdot(Z),
      
      n_start + static_cast<int>(Tprofile_vertical[floor][1]*n_total),
      n_start + static_cast<int>(Tprofile_vertical[floor][2]*n_total),
      
      n_start + static_cast<int>(Tprofile_vertical[floor][3]*n_total),
      Leg_ini.P(Z) + step_height * XPmiddle_ratio[floor] + SWmiddle_point[floor],
      0.0,
      0.0
      );
  }
  else if( n_period == static_cast<int>((1.0-Tprofile_vertical[floor][4])*n_total) ){
    SwingLegPos[Z].setMotionData(
      n_start,
      Leg_ini.P(Z),
      Leg_ini.V(Z),
      Leg_ini.Vdot(Z),
      
      n_start + static_cast<int>((Tprofile_vertical[floor][5]-Tprofile_vertical[floor][4])*n_total),
      n_start + static_cast<int>((Tprofile_vertical[floor][6]-Tprofile_vertical[floor][4])*n_total),
      
      n_start + static_cast<int>((Tprofile_vertical[floor][7]-Tprofile_vertical[floor][4])*n_total),
      Leg_end.P(Z),
      Leg_end.V(Z),
      Leg_end.Vdot(Z)
      );
  }  
  
  // Horizontal motion
  if( n_period + static_cast<int>(Tprofile_horizontal[floor][0]*n_total) >= n_total ){
    for( int i = X ; i <= Y ; i++ ){
      SwingLegPos[i].setMotionData(
        n_start + static_cast<int>(Tprofile_horizontal[floor][0]*n_total),
	Leg_ini.P(i),
	Leg_ini.V(i),
	Leg_ini.Vdot(i),
	
	n_start + static_cast<int>(Tprofile_horizontal[floor][1]*n_total),
	n_start + static_cast<int>(Tprofile_horizontal[floor][2]*n_total),
	
	n_start + static_cast<int>(Tprofile_horizontal[floor][3]*n_total),
	Leg_end.P(i),
	Leg_end.V(i),
	Leg_end.Vdot(i)
      );
    }
  }
  else{
    for( int i = X ; i <= Y ; i++ ){
      SwingLegPos[i].setMotionData(
        n_start,
	Leg_ini.P(i),
	Leg_ini.V(i),
	Leg_ini.Vdot(i),
	
	n_start + static_cast<int>(Tprofile_horizontal[floor][1]*n_total) + n_period - n_total,
	n_start + static_cast<int>(Tprofile_horizontal[floor][2]*n_total) + n_period - n_total,
	
	n_start + static_cast<int>(Tprofile_horizontal[floor][3]*n_total) + n_period - n_total,
	Leg_end.P(i),
	Leg_end.V(i),
	Leg_end.Vdot(i)
      );
    }
  }
}

void FootMotionGenerator::setSwingVerticalPosParam(State& Leg_ini, State& Leg_end, int n_start, int n_period, int n_total)
{
  static int floor = FLAT;
  // Vertical motion
  if( n_period == n_total ){
    const double step_height = Leg_end.P(Z) - Leg_ini.P(Z);
    floor = selectFloor(step_height);
    
    SwingLegPos[Z].setMotionData(
      n_start + static_cast<int>(Tprofile_vertical[floor][0]*n_total),
      Leg_ini.P(Z),
      Leg_ini.V(Z),
      Leg_ini.Vdot(Z),
      
      n_start + static_cast<int>(Tprofile_vertical[floor][1]*n_total),
      n_start + static_cast<int>(Tprofile_vertical[floor][2]*n_total),
      
      n_start + static_cast<int>(Tprofile_vertical[floor][3]*n_total),
      Leg_ini.P(Z) + step_height * XPmiddle_ratio[floor] + SWmiddle_point[floor],
      0.0,
      0.0
      );
  }
  else if( n_period == static_cast<int>((1.0-Tprofile_vertical[floor][4])*n_total) ){
    SwingLegPos[Z].setMotionData(
      n_start,
      Leg_ini.P(Z),
      Leg_ini.V(Z),
      Leg_ini.Vdot(Z),
      
      n_start + static_cast<int>((Tprofile_vertical[floor][5]-Tprofile_vertical[floor][4])*n_total),
      n_start + static_cast<int>((Tprofile_vertical[floor][6]-Tprofile_vertical[floor][4])*n_total),
      
      n_start + static_cast<int>((Tprofile_vertical[floor][7]-Tprofile_vertical[floor][4])*n_total),
      Leg_end.P(Z),
      Leg_end.V(Z),
      Leg_end.Vdot(Z)
      );
  }  
}  

void FootMotionGenerator::setSwingHorizontalPosParam(State& Leg_ini, State& Leg_end, int n_start, int n_total)
{
  static int floor = FLAT;
  // Horizontal motion
  for( int i = X ; i <= Y ; i++ ){
    SwingLegPos[i].setMotionData(
      n_start,
      Leg_ini.P(i),
      Leg_ini.V(i),
      Leg_ini.Vdot(i),
      
      n_start + static_cast<int>(Tprofile_horizontal[floor][1]*n_total),
      n_start + static_cast<int>(Tprofile_horizontal[floor][2]*n_total),
      
      n_start + static_cast<int>(Tprofile_horizontal[floor][3]*n_total),
      Leg_end.P(i),
      Leg_end.V(i),
      Leg_end.Vdot(i)
    );
  }
}

void FootMotionGenerator::setPivotParam(MotionInterpolator2& SupLegPivot_, MotionInterpolator2& SwgLegPivot_, vector3& step, WalkingState& ws, int n_start, int n_period)
{
  int SupportLEG = getSupportLeg(ws);
  
  const int floor = selectFloor(step(Z));
  const int NP_pivot0 = static_cast<int>(TP_pivot[floor][0]*n_period);
  const int NP_pivot1 = static_cast<int>(TP_pivot[floor][1]*n_period);
  
  if( step(X) >= 0.0 ){
    SwgLegPivot_[X].setMotionData(n_start+NP_pivot0,  foot_size[FB][TOE],  n_start+NP_pivot1, -foot_size[FB][HEEL]);
    SupLegPivot_[X].setMotionData(n_start,           -foot_size[FB][HEEL], n_start+n_period,   foot_size[FB][TOE]);
  }
  else{
    if( R_blandoff[floor] >= 0.0 ){
      SwgLegPivot_[X].setMotionData(n_start+NP_pivot0,  foot_size[FB][TOE],  n_start+NP_pivot1,  foot_size[FB][TOE]);
      SupLegPivot_[X].setMotionData(n_start,            foot_size[FB][TOE],  n_start+n_period,   foot_size[FB][TOE]);
    }
    else{
      SwgLegPivot_[X].setMotionData(n_start+NP_pivot0, -foot_size[FB][HEEL], n_start+NP_pivot1,  foot_size[FB][TOE]);
      SupLegPivot_[X].setMotionData(n_start,            foot_size[FB][TOE],  n_start+n_period,  -foot_size[FB][HEEL]);
    }
  }
  
  if( step(Y) >= 0.0 ){
    int inside  = SupportLEG==RLEG ? INSIDE : OUTSIDE;
    int outside = SupportLEG==RLEG ? OUTSIDE : INSIDE;
    SwgLegPivot_[Y].setMotionData(n_start+NP_pivot0, -foot_size[RL][outside], n_start+NP_pivot1,  foot_size[RL][inside]);
    SupLegPivot_[Y].setMotionData(n_start,            foot_size[RL][inside],  n_start+n_period,  -foot_size[RL][outside]);
  }
  else{
    int inside  = SupportLEG==RLEG ? INSIDE : OUTSIDE;
    int outside = SupportLEG==RLEG ? OUTSIDE : INSIDE;
    SwgLegPivot_[Y].setMotionData(n_start+NP_pivot0,  foot_size[RL][inside],  n_start+NP_pivot1, -foot_size[RL][outside]);
    SupLegPivot_[Y].setMotionData(n_start,           -foot_size[RL][outside], n_start+n_period,   foot_size[RL][inside]);
  }
}

void FootMotionGenerator::setSwingRpyParam(MotionInterpolator2& SwingRpyDat, vector3& step, double step_yaw, unsigned int FQsize, int n_start, int n_period)
{
  const int floor = selectFloor(step(Z));
  
  if( fabs(step_yaw) <= ToRad(5.0) && FQsize >= 1 ){
    vector3 R_landoff_mod(0.0), R_touchdown_mod(0.0);
    if( fabs(step(Y)) >= swg_rpy_threshold ){
      R_landoff_mod(X)   = -Sign(step(Y)) * R_landoff[floor](X);
      R_touchdown_mod(X) = -Sign(step(Y)) * R_touchdown[floor](X);
    }
    else{
      R_landoff_mod(X)   = -step(Y) * R_landoff[floor](X)   / swg_rpy_threshold;
      R_touchdown_mod(X) = -step(Y) * R_touchdown[floor](X) / swg_rpy_threshold;
    }
    
    if( step(X) >= 0 ){
      if( step(X) >= swg_rpy_threshold ){
	R_landoff_mod(Y)   = R_landoff[floor](Y);
	R_touchdown_mod(Y) = R_touchdown[floor](Y);
      }
      else{
	R_landoff_mod(Y)   = step(X) * R_landoff[floor](Y)   / swg_rpy_threshold;
	R_touchdown_mod(Y) = step(X) * R_touchdown[floor](Y) / swg_rpy_threshold;	
      }
    }
    else{
      if( step(X) <= -swg_rpy_threshold ){
	R_landoff_mod(Y)   = R_blandoff[floor];
	R_touchdown_mod(Y) = R_btouchdown[floor];
      }
      else{
	R_landoff_mod(Y)   = -step(X) * R_blandoff[floor]   / swg_rpy_threshold;
	R_touchdown_mod(Y) = -step(X) * R_btouchdown[floor] / swg_rpy_threshold;
      }
    }
    
    double angle_ratio = (double)(Nss + Nds) / (double)(n_period + Nds);
    
    vector<int> nd;
    for( int i = 0 ; i < 4 ; i++ )
      nd.push_back(n_start + static_cast<int>(TR_ankle[floor][i]*n_period));
    
    for( int i = 0 ; i < 2 ; i++ ){
      vector<double> pd;
      pd.push_back(0.0);
      pd.push_back(R_landoff_mod(i)   * angle_ratio);
      pd.push_back(R_touchdown_mod(i) * angle_ratio);
      pd.push_back(0.0);
      SwingRpyDat[i].setMotionData(nd, pd);
    }
  }
  else if( n_period > 1 ){
    vector<int> nd;
    vector<double> pd;
    for( int i = 0 ; i < 4 ; i++ ){
      nd.push_back(n_start + static_cast<int>(TR_ankle[floor][i]*n_period));
      pd.push_back(0.0);
    }
    for( int i = 0 ; i < 2 ; i++ )
      SwingRpyDat[i].setMotionData(nd, pd);
  }
}

void FootMotionGenerator::setAnkleParam(MotionInterpolator3& AnkleRpyDat, State& Leg_ini, State& Leg_end, int n_start, int n_period)
{
  for( int i = 0 ; i < 3 ; i++ ){
    AnkleRpyDat[i].setMotionData(
      n_start,
      Leg_ini.R_rpy(i),
      Leg_ini.W(i),
      Leg_ini.Wdot(i),
      n_start+n_period,
      Leg_end.R_rpy(i),
      Leg_end.W(i),
      Leg_end.Wdot(i)
    );
  }
}

#ifdef WITH_TOE_JOINT
void FootMotionGenerator::setToeParam(MotionInterpolator& SupToe, MotionInterpolator& SwgToe, unsigned int FQsize, int n_start, int n_period)
{
  double SupToe_ini, SwgToe_ini;
  SupToe.getMotionData(n_start, SupToe_ini);
  SwgToe.getMotionData(n_start, SwgToe_ini);
  
  if( FQsize >= 2 ){
    SupToe.setMotionData(
      n_start + NRtoe_sup(n_period, 0),
      SupToe_ini,
      n_start + NRtoe_sup(n_period, 1),
      TOE_PITCH_MAX * TOE_PITCH_RATIO
    );
  }
  else{
    SupToe.setMotionData(
      n_start + NRtoe_sup(n_period, 0),
      SupToe_ini,
      n_start + NRtoe_sup(n_period, 1),
      0.0
    );
  }

  SwgToe.setMotionData(
    n_start + NRtoe_swg(n_period, 0),
    SwgToe_ini,
    n_start + NRtoe_swg(n_period, 1),
    0.0
  );
}
#endif

#ifndef WITH_TOE_JOINT
// swing_rpy , pivotΰ֤ˤž(Υ,)
// foot_.R_rpy , ­ˤž()
void FootMotionGenerator::getAnkleMotion(MotionInterpolator2& PivotDat, MotionInterpolator2& SwingRpyDat, State& foot_, State& ankle_, vector3& pivot_, int n_clock)
{
  for( int i = 0 ; i < 2 ; i++ )
    PivotDat[i].getMotionData(n_clock, pivot_(i));
  
  vector3 swing_rpy(0.0), vswing_rpy(0.0), vswingdot_rpy(0.0);
  for( int i = 0 ; i < 2 ; i++ )
    SwingRpyDat[i].getMotionData(n_clock, swing_rpy(i), vswing_rpy(i), vswingdot_rpy(i));
  
  matrix33 Rpivot_f(rotFromRpy(swing_rpy));
  vector3 AnkleToPivot_f( Rpivot_f * (pivot_ - FootToAnkle) );

  matrix33 Rfoot( rotFromRpy(foot_.R_rpy) );
  
  vector3 x_cmd_pre( ankle_.P );
  vector3 v_cmd_pre( ankle_.V );
  ankle_.P    = foot_.P    + Rfoot * (pivot_ - AnkleToPivot_f );
  ankle_.V    = (ankle_.P - x_cmd_pre) / DT;
  ankle_.Vdot = (ankle_.V - v_cmd_pre) / DT;
  
  ankle_.R_rpy = foot_.R_rpy + swing_rpy;
  ankle_.R     = rotFromRpy(ankle_.R_rpy);
  ankle_.W     = foot_.W     + omegaFromRpyAngularVelocity(swing_rpy, vswing_rpy);
  ankle_.Wdot  = foot_.Wdot  + omegaFromRpyAngularVelocity(swing_rpy, vswingdot_rpy);
}

#else

// toe_rpy , ἴˤž
// swing_rpy , pivotΰ֤ˤž(Υ,)
// foot_.R_rpy , ­ˤž()
void FootMotionGenerator::getAnkleMotion(MotionInterpolator2& PivotDat, MotionInterpolator2& SwingRpyDat, MotionInterpolator& ToePitch, State& foot_, State& ankle_, double& toe_pitch_cmd, vector3& pivot_, int n_clock)
{
  double vtoe_pitch_cmd, vtoedot_pitch_cmd;
  ToePitch.getMotionData(n_clock, toe_pitch_cmd, vtoe_pitch_cmd, vtoedot_pitch_cmd);

  vector3 toe_rpy(0.0, -toe_pitch_cmd, 0.0);
  vector3 vtoe_rpy(0.0, -vtoe_pitch_cmd, 0.0);
  vector3 vtoedot_rpy(0.0, -vtoedot_pitch_cmd, 0.0);

  const vector3 AnkleToToe(FootToToe - FootToAnkle);
  matrix33 Rtoe_t = rotFromRpy(toe_rpy(X), toe_rpy(Y), toe_rpy(Z));
  vector3 AnkleToToe_t(Rtoe_t * AnkleToToe);
  vector3 FootToAnkle_t(FootToToe - AnkleToToe_t);
  
  for( int i = 0 ; i < 2 ; i++ )
    PivotDat[i].getMotionData(n_clock, pivot_(i));

  vector3 swing_rpy(0.0), vswing_rpy(0.0), vswingdot_rpy(0.0);
  for( int i = 0 ; i < 2 ; i++ )
    SwingRpyDat[i].getMotionData(n_clock, swing_rpy(i), vswing_rpy(i), vswingdot_rpy(i));
  
  matrix33 Rpivot_f(rotFromRpy(swing_rpy));
  vector3 AnkleToPivot_f( Rpivot_f * (pivot_ - FootToAnkle_t) );
  
  matrix33 Rfoot( rotFromRpy(foot_.R_rpy) );

  vector3 x_cmd_pre( ankle_.P );
  vector3 v_cmd_pre( ankle_.V );
  ankle_.P    = foot_.P    + Rfoot * (pivot_ - AnkleToPivot_f );
  ankle_.V    = (ankle_.P - x_cmd_pre) / DT;
  ankle_.Vdot = (ankle_.V - v_cmd_pre) / DT;
  
  ankle_.R_rpy = foot_.R_rpy + swing_rpy + toe_rpy;
  ankle_.R     = rotFromRpy(ankle_.R_rpy);
  ankle_.W     = foot_.W     + omegaFromRpyAngularVelocity(swing_rpy, vswing_rpy)    + vtoe_rpy;
  ankle_.Wdot  = foot_.Wdot  + omegaFromRpyAngularVelocity(swing_rpy, vswingdot_rpy) + vtoedot_rpy;
}
#endif


#ifndef WITH_TOE_JOINT
void FootMotionGenerator::makeSwingLegMotion(State Foot_cmd[2], State Ankle_cmd[2], vector3 Pivot_cmd[2], WalkingState ws, int n_clock)
{
  for( int i = 0 ; i < 3 ; i++ ){
    AnkleRpy[RLEG][i].getMotionData(n_clock, Foot_cmd[RLEG].R_rpy(i), Foot_cmd[RLEG].W(i), Foot_cmd[RLEG].Wdot(i));
    AnkleRpy[LLEG][i].getMotionData(n_clock, Foot_cmd[LLEG].R_rpy(i), Foot_cmd[LLEG].W(i), Foot_cmd[LLEG].Wdot(i));
  }
  
  for( int i = 0 ; i < 3 ; i++ ){
    if( isLeftSupport(ws) )
      SwingLegPos[i].getMotionData(n_clock, Foot_cmd[RLEG].P(i), Foot_cmd[RLEG].V(i), Foot_cmd[RLEG].Vdot(i));
    else if( isRightSupport(ws) )
      SwingLegPos[i].getMotionData(n_clock, Foot_cmd[LLEG].P(i), Foot_cmd[LLEG].V(i), Foot_cmd[LLEG].Vdot(i));
  }
  

  getAnkleMotion(Pivot[RLEG], SwingRpy[RLEG], Foot_cmd[RLEG], Ankle_cmd[RLEG], Pivot_cmd[RLEG], n_clock);
  getAnkleMotion(Pivot[LLEG], SwingRpy[LLEG], Foot_cmd[LLEG], Ankle_cmd[LLEG], Pivot_cmd[LLEG], n_clock);
}
#else
void FootMotionGenerator::makeSwingLegMotion(State *Foot_cmd, State *Ankle_cmd, double& toe_angle_right, double& toe_angle_left, vector3 *Pivot_cmd, WalkingState ws, int n_clock)
{
  for( int i = 0 ; i < 3 ; i++ ){
    AnkleRpy[RLEG][i].getMotionData(n_clock, Foot_cmd[RLEG].R_rpy(i), Foot_cmd[RLEG].W(i), Foot_cmd[RLEG].Wdot(i));
    AnkleRpy[LLEG][i].getMotionData(n_clock, Foot_cmd[LLEG].R_rpy(i), Foot_cmd[LLEG].W(i), Foot_cmd[LLEG].Wdot(i));
  }
  
  for( int i = 0 ; i < 3 ; i++ ){
    if( isLeftSupport(ws) )
      SwingLegPos[i].getMotionData(n_clock, Foot_cmd[RLEG].P(i), Foot_cmd[RLEG].V(i), Foot_cmd[RLEG].Vdot(i));
    else if( isRightSupport(ws) )
      SwingLegPos[i].getMotionData(n_clock, Foot_cmd[LLEG].P(i), Foot_cmd[LLEG].V(i), Foot_cmd[LLEG].Vdot(i));
  }
  
  getAnkleMotion(Pivot[RLEG], SwingRpy[RLEG], ToePitch[RLEG], Foot_cmd[RLEG], Ankle_cmd[RLEG], toe_angle_right, Pivot_cmd[RLEG], n_clock);
  getAnkleMotion(Pivot[LLEG], SwingRpy[LLEG], ToePitch[LLEG], Foot_cmd[LLEG], Ankle_cmd[LLEG], toe_angle_left,  Pivot_cmd[LLEG], n_clock);
}
#endif
