/*
 * Copyright (c) 2009,
 * @author Mitsuharu Morisawa
 *
 * AIST
 *
 * All rights reserved.
 *
 * This program is made available under the terms of the Eclipse Public License
 * v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 */
#include "FootStepManager.h"

using namespace mwalk_pattern_generator;


FootStepManager::FootStepManager(double Foot_distance_)
{
  Foot_distance = Foot_distance_;
  // default step restriction
  first_step = -1;
  STEP_X_FORWARD_MAX =  0.25;
  STEP_X_BACK_MAX    = -0.2;
  STEP_Y_OUT_MAX     =  0.15;
  STEP_Y_IN_MAX      =  0.05;
  STEP_Z_UP_MAX      =  0.10;
  STEP_Z_DOWN_MAX    = -0.10;
  STEP_YAW_OUT_MAX   = ToRad(30.0);
  STEP_YAW_IN_MAX    = ToRad(10.0);
}

FootStepManager::~FootStepManager()
{
}

bool FootStepManager::pushFootPlace(deque<FootPlace>& FootQueue_,
				    int support_leg_, double x, double y,
				    double z, double th, double hip_z, double roll,
				    double pitch, double yaw, double tss,
				    double tds, unsigned int uid)
{
  FootPlace FP;
  bool isModified = false;

  FP.uid = uid;
  FP.p = x,y,z;
  FP.rpy = 0.0, 0.0, th;
  FP.hip_z = hip_z;
  FP.frpy = roll, pitch, yaw;

  if( tss <= 0.0 ){
    FP.nss = Nss;
    isModified = true;
  }
  else
    FP.nss = (int)round(tss/DT);

  if( tds <= 0.0 ){
    FP.nds = Nds;
    isModified = true;
  }
  else
    FP.nds = (int)round(tds/DT);

  if( support_leg_ != -1 )
    FP.support_leg = support_leg_;
  else if( !FootQueue_.empty() )
    FP.support_leg = FootQueue_.back().support_leg == RLEG ? LLEG : RLEG;
  else if( first_step != -1 )
    FP.support_leg = first_step == RLEG ? LLEG : RLEG;
  else{
#ifdef ORIGIN_PATH
    if( FP.p(Y) < 0.0 || FP.rpy(Z) < 0.0 )
      FP.support_leg = LLEG;
    else
      FP.support_leg = RLEG;
#else
    FP.support_leg = LLEG;
#endif
  }

  if( FP.p(X) > STEP_X_FORWARD_MAX ){
    FP.p(X) = STEP_X_FORWARD_MAX;
    isModified = true;
  }
  else if( FP.p(X) < STEP_X_BACK_MAX ){
    FP.p(X) = STEP_X_BACK_MAX;
    isModified = true;
  }

#ifdef ORIGIN_PATH
  if( FP.support_leg == RLEG ){
    if( FP.p(Y) > STEP_Y_OUT_MAX ){
      FP.p(Y) =  STEP_Y_OUT_MAX;
      isModified = true;
    }
    else if( FP.p(Y) < -STEP_Y_IN_MAX ){
      FP.p(Y) = -STEP_Y_IN_MAX;
      isModified = true;
    }

    if( FP.rpy(2) > STEP_YAW_OUT_MAX ){
      FP.rpy(2) =  STEP_Y_OUT_MAX;
      isModified = true;
    }
    else if( FP.rpy(2) < -STEP_YAW_IN_MAX ){
      FP.rpy(2) = -STEP_YAW_IN_MAX;
      isModified = true;
    }
  }
  else if( FP.support_leg == LLEG ){
    if( FP.p(Y) > STEP_Y_IN_MAX ){
      FP.p(Y) =  STEP_Y_IN_MAX;
      isModified = true;
    }
    else if( FP.p(Y) < -STEP_Y_OUT_MAX ){
      FP.p(Y) = -STEP_Y_OUT_MAX;
      isModified = true;
    }

    if( FP.rpy(2) > STEP_YAW_IN_MAX ){
      FP.rpy(2) =  STEP_YAW_IN_MAX;
      isModified = true;
    }
    else if( FP.rpy(2) < -STEP_YAW_OUT_MAX ){
      FP.rpy(2) = -STEP_YAW_OUT_MAX;
      isModified = true;
    }
  }
#endif

  if( FP.p(Z) > STEP_Z_UP_MAX ){
    FP.p(Z) = STEP_Z_UP_MAX;
    isModified = true;
  }
  else if( FP.p(Z) < STEP_Z_DOWN_MAX ){
    FP.p(Z) = STEP_Z_DOWN_MAX;
    isModified = true;
  }

  FootQueue_.push_back(FP);

  return isModified;
}

void
FootStepManager::pushFootPlaceNoCheck(deque<FootPlace>& FootQueue_,
				      int support_leg_, double x, double y,
				      double z, double th, double hip_z,
				      double roll, double pitch, double yaw,
				      double tss, double tds, unsigned int uid)
{
  FootPlace FP;

  FP.uid = uid;
  FP.p = x,y,z;
  FP.rpy = 0.0, 0.0, th;
  FP.hip_z = hip_z;
  FP.frpy = roll, pitch, yaw;

  FP.nss = (int)round(tss/DT);
  FP.nds = (int)round(tds/DT);

  if( support_leg_ != -1 )
    FP.support_leg = support_leg_;
  else if( !FootQueue_.empty() )
    FP.support_leg = FootQueue_.back().support_leg == RLEG ? LLEG : RLEG;
  else if( first_step != -1 )
    FP.support_leg = first_step == RLEG ? LLEG : RLEG;
  else{
#ifdef ORIGIN_PATH
    if( FP.p(Y) < 0.0 || FP.rpy(Z) < 0.0 )
      FP.support_leg = LLEG;
    else
      FP.support_leg = RLEG;
#else
    FP.support_leg = LLEG;
#endif
  }
  FootQueue_.push_back(FP);
}
