/*
 * Copyright (c) 2009,
 * @author Mitsuharu Morisawa
 *
 * AIST
 *
 * All rights reserved.
 *
 * This program is made available under the terms of the Eclipse Public License
 * v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 */
#ifndef __LINEAR_CONTROL_H__
#define __LINEAR_CONTROL_H__

#include <ublasCommonTypes.h>

using namespace OpenHRP;

namespace linear_control_system{
  
  // linear control for siso system
  class LinearControl_SISO{
  public:
    // system matrices and feedback gain
    dmatrix A, B, C, K;
    
    dvector x;            // state variable
    double  y;            // output
    double  u;            // input
    
    void Initialize(void);
    LinearControl_SISO(void);
    virtual ~LinearControl_SISO(void);
    
    virtual void update(void);
    void setExtendedSystem(dmatrix& Ax, dmatrix& Bx, dmatrix& Cx);
    bool solveDiscreteLQRegulator(dmatrix& K, dmatrix& P, dmatrix& Ad, dmatrix& Bd, dmatrix& Cd, dmatrix& Qd, double Rd);
    inline bool solveDiscreteLQRegulator(dmatrix& K, dmatrix& P, dmatrix& Ad, dmatrix& Bd, dmatrix& Cd, double Qd, double Rd){
      dmatrix CtQC(ublas::prod(ublas::trans(Cd)*Qd, Cd));
      return solveDiscreteLQRegulator(K, P, Ad, Bd, Cd, CtQC, Rd);
    };
    inline bool solveDiscreteLQRegulator(dmatrix& P, dmatrix& Qd, double Rd){
      return solveDiscreteLQRegulator(K, P, A, B, C, Qd, Rd);
    };
    inline bool solveDiscreteLQRegulator(dmatrix& P, double Qd, double Rd){
      dmatrix CtQC(ublas::prod(ublas::trans(C)*Qd, C));
      return solveDiscreteLQRegulator(K, P, A, B, C, CtQC, Rd);
    };
    inline bool solveDiscreteLQRegulator(dmatrix& Qd, double Rd){
      dmatrix P(didentity(A.size1()));
      return solveDiscreteLQRegulator(K, P, A, B, C, Qd, Rd);
    };
  };
  
}; // end of namespace linear_control_system

#endif
