/*
 * Copyright (c) 2009,
 * @author Mitsuharu Morisawa
 *
 * AIST
 *
 * All rights reserved.
 *
 * This program is made available under the terms of the Eclipse Public License
 * v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 */
#include "LinearControl3.h"

using namespace linear_control_system;
using namespace tvmet;


LinearControl3::LinearControl3(void)
{
  LinearControl3::Initialize();
}

LinearControl3::~LinearControl3()
{
}

void LinearControl3::Initialize(void)
{
  A = 0;
  B = 0;
  C = 0;
  K = 0;
  x = 0;
  y = 0;
  u = 0;
}

void LinearControl3::update(void)
{
  y = C(0,0)*x(0) * C(0,1)*x(1) + C(0,2)*x(2);
  alias(x) = A * x + B * u;
}

void LinearControl3::update(const double dt)
{
  y = C(0,0)*x(0) * C(0,1)*x(1) + C(0,2)*x(2);
  vector3 dx(A * x + B * u);
  x += dx * dt;
}

#if 0
bool LinearControl3::solveDiscreteLQRegulator(vector3& Kd, matrix33& P, matrix33& Ad, vector3& Bd, vector3_t& Cd, matrix33& Qd, double Rd)
{
  const static double Peps = 1.0e-15; // «
  const static int    Nmax = 10000;   // ׻
  
  // solve riccati equation
  matrix33 Ppre(0.0);
  vector3  ddiagP(0.0);
  
  int n = 0;
  
  do{
    Ppre =  P;
    
    matrix33  AtP(tvmet::trans(Ad) * P);
    vector3_t Bt(Bd(0), Bd(1), Bd(2));
    vector3_t BtP(Bt, P);
    matrix33  AtPA(AtP * Ad);
    vector3   AtPB(AtP * Bd);
    vector3_t BtPA(AtPB(0), AtPB(1), AtPB(2));
    matrix33  AtPBBtPA(AtPB, BtPA);
    double    BtPB = (BtP * Bd)(0);
    
    P = Qd + AtPA - AtPBBtPA / (Rd + BtPB);
    
    ddiagP =
      P(0,0) - Ppre(0,0),
      P(1,1) - Ppre(1,1),
      P(2,2) - Ppre(2,2);
    
    n++;
  }while( n < Nmax && tvmet::dot(ddiagP, ddiagP) > Peps );
  
  vector3_t Bt(Bd(0), Bd(1), Bd(2));
  vector3_t BtP(Bt * P);
  vector3_t BtPA(BtP * Ad);
  double    BtPB = (BtP * Bd)(0);
  
  Kd = BtPA / (Rd + BtPB(0,0));
  
  if( n == Nmax )
    return false;
  
  return true;
}
#endif
