/*
 * Copyright (c) 2009,
 * @author Mitsuharu Morisawa
 *
 * AIST
 *
 * All rights reserved.
 *
 * This program is made available under the terms of the Eclipse Public License
 * v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 */
#ifndef __MODEL_MANAGER_H__
#define __MODEL_MANAGER_H__

#include <iostream>
#include "hrpModelHeaders.h"
#include "bodyinfo.h"
#include "mphysics.h"

using namespace OpenHRP;
using namespace std;

namespace motion_generator{
  
#if defined (HRP2NF1) || defined(HRP4L) || defined(HRP4)
#define WITH_TOE_JOINT
#endif
  
#if defined(HRP2) || defined(HRP2JRL) || defined(HRP2NF1) || defined(HRP3) || defined (HRP4L) || defined (HRP4)
#define WITH_CHEST_JOINT
#endif
  
#if defined(HRP2) || defined(HRP2JRL) || defined(HRP2NF1) || defined(HRP3) || defined (HRP4)
#define WITH_ARM_JOINT
#endif
  
#if defined(HRP2) || defined(HRP2JRL) || defined(HRP2NF1) || defined(HRP3) || defined (HRP4)
#define WITH_NECK_JOINT
#endif
  
#if defined(HRP2) || defined(HRP2JRL) || defined (HRP2NF1)
#define MOSENSOR_ON_CHEST
#endif
  
#ifdef WITH_NECK_JOINT
#if defined(HRP2) || defined(HRP2JRL) || defined(HRP2NF1) || defined(HRP3)
#define NECK_JOINT HEAD_JOINT1
#elif defined(HRP4)
#define NECK_JOINT NECK_P
#endif
#endif

#ifndef WITH_TOE_JOINT
  const static int     DOF_LEG = 12;
#else
  const static int     DOF_LEG = 14;
#endif
  
#ifdef WITH_CHEST_JOINT
#if defined(HRP4L) || defined(HRP4)
  const static int     DOF_CHEST = 3;
#else
  const static int     DOF_CHEST = 2;
#endif
#endif
  
#ifdef WITH_ARM_JOINT
  const static int     DOF_ARM = 6;
#endif
  
#ifdef WITH_NECK_JOINT
#if defined(HRP4)
  const static int     DOF_NECK = 3;
#else
  const static int     DOF_NECK = 2;
#endif
#endif  
  
  const static int     attitude_sensor_id = 0;
  const static int     gyro_id = 0;
  const static double  foot_size[2][2] = FOOT_SIZE;
  const static vector3 FootToAnkle(0.0, 0.0, ANKLE_HEIGHT);
  
#ifdef WITH_TOE_JOINT
#if defined (HRP2NF1) || defined(HRP4L) || defined(HRP4)
  static const double TOE_HEIGHT = 0.016;
  static const vector3 FootToToe(0.115, 0.0, TOE_HEIGHT);
#endif
#endif // WITH_TOE_JOINT
  
  class ModelManager{
  protected:
    void setVelocityLimit(double PowerSupply);
  public:
    BodyPtr body;
    Link* waistLink;
    Link* ankleLink[2];
    Link* kneeLink[2];
    Link* hipLink[2];
    Link* wristLink[2];
    Link* chestLink;
    Link* neckLink;
    
    JointPathPtr waistToAnklePath[2];
    JointPathPtr waistToWristPath[2];
    JointPathPtr waistToChestPath;
    JointPathPtr waistToNeckPath;
    
    Sensor* gsensobj;
    
    double  Foot_distance;
    double  Leg_length_max;
    double  Total_mass;
    vector3 Xg_offset_ini;
    
    double pos_save[DOF];         // used to check position/velocity consistency
    double joint_max[DOF];        // maximum joint angle
    double joint_min[DOF];        // minimum joint angle
    double vjoint_max[DOF];       // maximum joint speed
    
    void setupLinkModel(void);
    void setupModel(istringstream& strm);
    void setupModel(int argc, char* argv[]); 
    void setupModel(const char *model, int argc, char* argv[]);
    
    void calcModelZMP(vector3& zmp_model_, double zmp_z, vector3& LinMom_pre, vector3& AngMom_pre, bool& firsttime);
    void calcModelZMP(vector3& zmp_model_, vector3& total_com_, double zmp_z, vector3& LinMom_pre, vector3& AngMom_pre, bool& firsttime);
    void calcGsensorReference(vector3& gsens, vector3& o_Vgses_pre, bool& firsttime);
    
    void checkJointLimits(double t_now, bool& firsttime);
    void showJointLimits(void);
    
    void Initialize(void);
    
    ModelManager(istringstream& strm);
    ModelManager(int argc, char* argv[]);
    ModelManager(const char *model, int argc, char* argv[]);
    ModelManager(BodyPtr& body_);
    ~ModelManager();
  };

};  // end of namespace motion_generator


#endif // __MODEL_MANAGER_H__
