/*
 * Copyright (c) 2009,
 * @author Mitsuharu Morisawa
 *
 * AIST
 *
 * All rights reserved.
 *
 * This program is made available under the terms of the Eclipse Public License
 * v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 */
#include "MwalkFootPlanner.h"

using namespace foot_planner;


MwalkFootPlanner::MwalkFootPlanner(FootStepManager *fsm_)
{
  fsm = fsm_;
}

MwalkFootPlanner::~MwalkFootPlanner()
{
}

void MwalkFootPlanner::setOmniDirectionalWalk(deque<FootPlace>& FootQueue, double dx_cmd, double dy_cmd, double dyaw_cmd, int nss, int nds, bool verbose)
{
  vector3 FPi(0.0);
  int nx = 0;
  if( dx_cmd > 0.0 ){
    nx = (int)(dx_cmd / fsm->STEP_X_FORWARD_MAX + 1.0e-10);
    
    double firststep_x = dx_cmd - (double)(nx * fsm->STEP_X_FORWARD_MAX);
    if( firststep_x > 1.0e-5){
      FPi(X) = (firststep_x + fsm->STEP_X_FORWARD_MAX) / 2.0;
      nx++;
    }
    else if(nx > 0)
      FPi(X) = fsm->STEP_X_FORWARD_MAX;
  }
  else if( dx_cmd < 0.0){
    nx = (int)(dx_cmd / fsm->STEP_X_BACK_MAX + 1.0e-10);
    
    double firststep_x = dx_cmd - (double)(nx * fsm->STEP_X_BACK_MAX);
    if( firststep_x < -1.0e-5 ){
      FPi(X) = (firststep_x + fsm->STEP_X_BACK_MAX) / 2.0;
      nx++;
    }
    else if(nx > 0)
      FPi(X) = fsm->STEP_X_BACK_MAX;
  }
  else
    FPi(X) = 0.0;
  
  int ny = 0;
  if( dy_cmd > 0.0 ){
    ny = (int)(dy_cmd / fsm->STEP_Y_OUT_MAX + 1.0e-6) * 2;
    
    double firststep_y = dy_cmd - (double)((ny/2) * fsm->STEP_Y_OUT_MAX);
    if( firststep_y > 1.0e-5 ){
      FPi(Y) = (firststep_y + fsm->STEP_Y_OUT_MAX) / 2.0;
      ny+=2;
    }
    else if( ny > 0 )
      FPi(Y) = fsm->STEP_Y_OUT_MAX;
    else
      FPi(Y) = 0.0;
  }
  else{
    ny = (int)(-dy_cmd / fsm->STEP_Y_OUT_MAX + 1.0e-6) * 2;
    
    double firststep_y = dy_cmd + (double)((ny / 2)* fsm->STEP_Y_OUT_MAX);
    if( firststep_y < -1.0e-5 ){
      FPi(Y) = (firststep_y - fsm->STEP_Y_OUT_MAX) / 2.0;
      ny+=2;
    }
    else if( ny > 0 )
      FPi(Y) = -fsm->STEP_Y_OUT_MAX;
    else
      FPi(Y) = 0.0;
  }
  
  int nr = 0;
  if( dyaw_cmd > 0.0 ){
    nr = (int)(dyaw_cmd / fsm->STEP_YAW_OUT_MAX + 1.0e-6);
    
    double firststep_r = dyaw_cmd - (double)(nr * fsm->STEP_YAW_OUT_MAX);
    if( firststep_r > 1.0e-10 ){
      FPi(Z) = (firststep_r + fsm->STEP_YAW_OUT_MAX) / 2.0;
      nr++;
    }
    else if( nr > 0 )
      FPi(Z) = fsm->STEP_YAW_OUT_MAX;
    else
      FPi(Z) = 0.0;
  }
  else{
    nr = (int)(-dyaw_cmd / fsm->STEP_YAW_OUT_MAX + 1.0e-6);
    
    double firststep_r = dyaw_cmd + (double)(nr * fsm->STEP_YAW_OUT_MAX);
    if( firststep_r < -1.0e-10 ){
      FPi(Z) = (firststep_r - fsm->STEP_YAW_OUT_MAX) / 2.0;
      nr++;
    }
    else if( nr > 0 )
      FPi(Z) = -fsm->STEP_YAW_OUT_MAX;
    else
      FPi(Z) = 0.0;
  }
  
  int n = 0;
  double dyaw_total = 0.0;
  
  while( nx > n || ny > n || nr > n){
    vector3 dp(0.0);
    double  dyaw = 0.0;
    if( nx >= ny && nx >= nr ){
      if( n == 0 || nx - 1 == n )
	dp(X) = FPi(X);
      
      else if( nx - 1 > n ){
	if( dx_cmd > 0.0 )
	  dp(X) = fsm->STEP_X_FORWARD_MAX;
	else if( dx_cmd < 0.0 )
	  dp(X) = fsm->STEP_X_BACK_MAX;
	else
	  dp(X) = 0.0;
      }
      
      else
	dp(X) = 0.0;
      
      if( n % 2 == 0 )
	dp(Y) = dy_cmd / (double)(nx);
      else
	dp(Y) = 0.0;
      dyaw = dyaw_cmd / (double)(nx);
    }
    else if( ny >= nx && ny >= nr ){
      if( n == 0 || ny - 2 == n )
	dp(Y) = FPi(Y);
      
      else if( ny - 2 > n ){
	if( dy_cmd > 0.0 ){
	  if( n % 2 == 0 )
	    dp(Y) = fsm->STEP_Y_OUT_MAX;
	  else
	    dp(Y) = 0.0;
	}
	else if( dy_cmd < 0.0 ){
	  if( n % 2 == 0 )
	    dp(Y) = -fsm->STEP_Y_OUT_MAX;
	  else
	    dp(Y) = 0.0;
	}
	else
	  dp(Y) = 0.0;
      }
      
      else
	dp(Y) = 0.0;
      
      dp(X) = dx_cmd / (double)(ny);
      dyaw = dyaw_cmd / (double)(ny);
    }
    else{
      if( n == 0 || nr - 1 == n )
	dyaw = FPi(Z);
      
      else if( nr - 1 > n ){
	if( dyaw_cmd > 0.0 )
	  dyaw =  fsm->STEP_YAW_OUT_MAX;
	else if( dyaw_cmd < 0.0 )
	  dyaw = -fsm->STEP_YAW_OUT_MAX;
	else
	  dyaw = 0.0;
      }
      
      else
	dyaw = 0.0;
      
      dp(X) = dx_cmd / (double)(nr);
      if( n % 2 == 0 )
	dp(Y) = dy_cmd / (double)(nr);
      else
	dp(Y) = 0.0;
    }

    FootPlace FP;    
    if( !FootQueue.empty() )
      FP.support_leg = FootQueue.back().support_leg == RLEG ? LLEG : RLEG;
    else if( fsm->first_step != -1 )
      FP.support_leg = fsm->first_step;
    else if( FP.p(Y) < 0.0 || FP.rpy(Z) < 0.0 )
      FP.support_leg = LLEG;
    else
      FP.support_leg = RLEG;
    
    // convert to support foot coordinates
    const double foot_distance_half = (FP.support_leg == RLEG ) ? -fsm->Foot_distance/2.0 : fsm->Foot_distance/2.0;
    const double dx =  foot_distance_half * sin(dyaw_total);
    const double dy =  foot_distance_half * (1.0 - cos(dyaw_total));
    
    FP.p(X) =  dp(X) * cos(dyaw_total) + dp(Y) * sin(dyaw_total) - dx;
    FP.p(Y) = -dp(X) * sin(dyaw_total) + dp(Y) * cos(dyaw_total) - dy;
    FP.p(Z) =  0.0;
    FP.rpy(X) = 0.0;
    FP.rpy(Y) = 0.0;
    FP.rpy(Z) = dyaw;
    FP.nss = nss;
    FP.nds = nds;
    
    dyaw_total += dyaw;
    
    FootQueue.push_back(FP);
    n++;
  }
  
  if( verbose ){
    for( int i = 0 ; i < (int)FootQueue.size() ; i++ )
      cout << "FootQueue(x,y,th) = " << FootQueue[i].p(X) << "," << FootQueue[i].p(Y) << "," << ToDeg(FootQueue[i].rpy(Z)) << ", nss = " << nss << ", nds = " << nds << endl;
  }
}

// import from setTargetPosNoWait() in hwalk.cpp
void MwalkFootPlanner::setTargetPos(deque<FootPlace>& FootQueue, double dx_cmd, double dy_cmd, double dyaw_cmd, int nss, int nds, bool verbose)
{
  if( dx_cmd >= fsm->STEP_X_FORWARD_MAX || dx_cmd <= fsm->STEP_X_BACK_MAX || fabs(dy_cmd) >= fsm->STEP_Y_OUT_MAX ){
    double dyaw_target = atan2(dy_cmd, dx_cmd);
    
    if( (dx_cmd * dy_cmd > 0.0 && dyaw_target < dyaw_cmd) || (dx_cmd * dy_cmd < 0.0 && dyaw_target > dyaw_cmd) ){
      double ab = dy_cmd / tan(dyaw_cmd);
      double bs = 0;
      if (fabs(dyaw_cmd) == M_PI/2){
	bs = dx_cmd;
      }else{
	bs = dx_cmd - ab;
      }
      
      double be = dy_cmd / sin(dyaw_cmd);
      if( fabs(bs) <= fabs(be) ){
	double r = bs / tan(dyaw_cmd / 2.0);
	setArcWalk(FootQueue, 0.0, r, dyaw_cmd, nss, nds, verbose);
	setOmniDirectionalWalk(FootQueue, be-bs, 0.0, 0.0, nss, nds, verbose);
      }
      else{
	double r = be / tan(dyaw_cmd / 2.0);
	setOmniDirectionalWalk(FootQueue, bs-be, 0.0, 0.0, nss, nds, verbose);
	setArcWalk(FootQueue, 0, r, dyaw_cmd, nss, nds, verbose);
      }
    }
    else if( ((dx_cmd * dy_cmd > 0.0 && dyaw_cmd <= dyaw_target) || (dx_cmd * dy_cmd < 0.0 && dyaw_cmd >= dyaw_target)) && (fabs(dyaw_cmd) > fsm->STEP_YAW_OUT_MAX) ){
      setOmniDirectionalWalk(FootQueue, dx_cmd, dy_cmd, 0.0, nss, nds, verbose);
      setOmniDirectionalWalk(FootQueue, 0.0, 0.0, dyaw_cmd, nss, nds, verbose);
    }
    else{
      setOmniDirectionalWalk(FootQueue, dx_cmd, dy_cmd, dyaw_cmd, nss, nds, verbose);
    }
  }else{
    setOmniDirectionalWalk(FootQueue, dx_cmd, dy_cmd, dyaw_cmd, nss, nds, verbose);
  }
}

void MwalkFootPlanner::setArcWalk(deque<FootPlace>& FootQueue, double cx_cmd, double cy_cmd, double dyaw_cmd, int nss, int nds, bool verbose)
{
  
  if( fabs(dyaw_cmd) > 0.0 && (cx_cmd != 0.0 || cy_cmd != 0.0) ){
    double yaw_ini = atan2(-cy_cmd, -cx_cmd);
    if( cy_cmd > 0.0 )
      yaw_ini += M_PI / 2.0;
    else if( cy_cmd < 0.0 )
      yaw_ini -= M_PI / 2.0;
    else if( dyaw_cmd > 0.0 )
      yaw_ini += M_PI / 2.0;
    else
      yaw_ini -= M_PI / 2.0;
    double yaw_total = yaw_ini + dyaw_cmd;
    
    double arc_length = dyaw_cmd * sqrt(cx_cmd*cx_cmd+cy_cmd*cy_cmd);
    int nx = 0;
    if( yaw_total > 0.0 )
      nx = (int)ceil(fabs(arc_length * cos(dyaw_cmd) / fsm->STEP_X_FORWARD_MAX));
    else
      nx = (int)ceil(fabs(-arc_length * cos(dyaw_cmd) / fsm->STEP_X_BACK_MAX));
    int ny = (int)ceil(fabs(2.0 * arc_length * sin(dyaw_cmd) / (fsm->STEP_Y_OUT_MAX + fsm->STEP_Y_IN_MAX)));
    int nr = (int)ceil(fabs(yaw_total / fsm->STEP_YAW_OUT_MAX));
    
    int n_arc = Max(nr, Max(nx, ny));
    
    double cthi = cos(yaw_ini);
    double sthi = sin(yaw_ini);
    
    double yaw_step = yaw_ini + dyaw_cmd / (double)n_arc;
    double cth = cos(yaw_step);
    double sth = sin(yaw_step);
    
    vector3 dp((cth-cthi)*cx_cmd + (sth-sthi)*cy_cmd, -(sth-sthi)*cx_cmd + (cth-cthi)*cy_cmd, 0.0);
    double dyaw = yaw_total / (double)n_arc;
    
    for( int i = 0 ; i < n_arc ; i++ ){
      FootPlace FP;
      if( !FootQueue.empty() )
	FP.support_leg = (FootQueue.back().support_leg == RLEG) ? LLEG : RLEG;
      else if( fsm->first_step != -1 )
	FP.support_leg = fsm->first_step;
      //else if( FP.p(Y) <= 0.0 || FP.rpy(Z) <= 0.0 )
      else if( cx_cmd * dyaw_cmd > 0.0 || (cx_cmd == 0.0 && cy_cmd > 0.0) )
	FP.support_leg = RLEG;
      else
	FP.support_leg = LLEG;
      
#if 0
      // convert to support foot coordinates
      const double foot_distance_half = (FP.support_leg == RLEG ) ? -fsm->Foot_distance/2.0 : fsm->Foot_distance/2.0;
      const double dx =  foot_distance_half * sin(dyaw);
      const double dy =  foot_distance_half * (1.0 - cos(dyaw));
      
      FP.p(X) =  dp(X) * cos(dyaw) + dp(Y) * sin(dyaw) - dx;
      FP.p(Y) = -dp(X) * sin(dyaw) + dp(Y) * cos(dyaw) - dy;
#endif
      FP.p(X) =  dp(X);
      FP.p(Y) =  dp(Y);
      FP.p(Z) =  0.0;
      FP.rpy(X) = 0.0;
      FP.rpy(Y) = 0.0;
      FP.rpy(Z) = dyaw;
      FP.nss = nss;
      FP.nds = nds;
      
      FootQueue.push_back(FP);
    }
    
    if( verbose ){
      for( int i = 0 ; i < (int)FootQueue.size() ; i++ )
	cout << "FootQueue(x,y,th) = " << FootQueue[i].p(X) << "," << FootQueue[i].p(Y) << "," << ToDeg(FootQueue[i].rpy(Z)) << ", nss = " << nss << ", nds = " << nds << endl;
    }
  }
  else
    cout << "invalid command!!" << endl;
}

void MwalkFootPlanner::setTargetPosPassGate(deque<FootPlace>& FootQueue, vector3& gate_pos1, vector3& gate_pos2, vector3& target_pos, int nss, int nds, bool verbose)
{
  const double step_before_down_stage = 0.20;
  const double step_before_up_stage = 0.25;
  const double step_after_down_stage = 0.25;
  const double step_after_up_stage = 0.20;
  const double step_before_down_margin = 0.1;
  const double step_before_up_margin = 0.05;
  
  vector3 gate_mid((gate_pos1(X)+gate_pos2(X))/2.0, (gate_pos1(Y)+gate_pos2(Y))/2.0, 0.0);
  vector3 pass_dir(tvmet::normalize(vector3(gate_pos1(Y)-gate_pos2(Y), gate_pos1(X)-gate_pos2(X), 0.0)));
  if( gate_mid(X)*pass_dir(Y) - gate_mid(Y)*pass_dir(X) < 0.0 )
    pass_dir = -pass_dir;

  double stage_height = (gate_pos1(Z)+gate_pos2(Z)) / 2.0;
  double angle_pass = atan2(pass_dir(Y), pass_dir(X));
  
  vector3 middle_pos1, middle_pos2;
  if( stage_height > 0.0 ){
    middle_pos1 = gate_mid - pass_dir * ((step_before_up_stage + step_before_up_margin) + step_before_up_stage);
    middle_pos2 = gate_mid + pass_dir * step_after_up_stage;
    
#if 1
    double angle1 = atan2(middle_pos1(Y), middle_pos1(X));
    double length = sqrt(middle_pos1(X)*middle_pos1(X) + middle_pos1(Y)*middle_pos1(Y));
    if( angle1 != 0.0 )
      setTargetPos(FootQueue, 0.0, 0.0, angle1, nss, nds, verbose);
    if( length > 0.0 )
      setTargetPos(FootQueue, length, 0.0, 0.0, nss, nds, verbose);

    double angle12 = angle_pass - angle1;
    if( angle12 > M_PI ) angle12 -= 2.0*M_PI;
    else if( angle12 < -M_PI ) angle12 += 2.0*M_PI;
    if( angle12 != 0.0 )
      setTargetPos(FootQueue, 0.0, 0.0, angle12, nss, nds, verbose);
#else
    setTargetPos(FootQueue, middle_pos1(X), middle_pos1(Y), angle_pass, nss, nds, verbose);
#endif
    
    fsm->pushFootPlaceNoCheck(FootQueue, -1, step_before_up_stage, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, (double)nss*DT, (double)nds*DT);  
    fsm->pushFootPlaceNoCheck(FootQueue, -1, step_before_up_stage, 0.0, stage_height, 0.0, 0.0, 0.0, 0.0, 0.0, (double)nss*DT, (double)nds*DT);  
    fsm->pushFootPlaceNoCheck(FootQueue, -1, step_after_up_stage,  0.0, 0.0, 0.0, stage_height, 0.0, 0.0, 0.0, (double)nss*DT, (double)nds*DT);  
    
#if 1
    double angle2 = atan2(target_pos(Y)-middle_pos2(Y), target_pos(X)-middle_pos2(X));
    double length2 = sqrt((target_pos(X)-middle_pos2(X))*(target_pos(X)-middle_pos2(X)) + (target_pos(Y)-middle_pos2(Y))*(target_pos(Y)-middle_pos2(Y))); 
    if( angle2 != 0.0 )
      setTargetPos(FootQueue, 0.0, 0.0, angle2-angle_pass, nss, nds, verbose);
    if( length2 > 0.0 )
      setTargetPos(FootQueue, length2, 0.0, 0.0, nss, nds, verbose);
    
    double angle22 = target_pos(Z) - angle2;
    if( angle22 > M_PI ) angle22 -= 2.0*M_PI;
    else if( angle22 < -M_PI ) angle22 += 2.0*M_PI;
    if( angle22 != 0.0 )
      setTargetPos(FootQueue, 0.0, 0.0, angle22, nss, nds, verbose);
#else
    // notice: need to convert coordinates
    setTargetPos(FootQueue, target_pos(X)-middle_pos2(X), target_pos(Y)-middle_pos2(Y), target_pos(Z)-angle_pass, nss, nds, verbose);    
#endif    
  }
  else if( stage_height < 0.0){
    middle_pos1 = gate_mid - pass_dir * (step_before_down_stage + step_before_down_margin);
    middle_pos2 = gate_mid + pass_dir * step_before_down_stage * 2.0;
    
#if 1
    double angle1 = atan2(middle_pos1(Y), middle_pos1(X));
    double length = sqrt(middle_pos1(X)*middle_pos1(X) + middle_pos1(Y)*middle_pos1(Y));
    if( angle1 != 0.0 )
      setTargetPos(FootQueue, 0.0, 0.0, angle1, nss, nds, verbose);
    if( length > 0.0 )
      setTargetPos(FootQueue, length, 0.0, 0.0, nss, nds, verbose);
    
    double angle12 = angle_pass - angle1;
    if( angle12 > M_PI ) angle12 -= 2.0*M_PI;
    else if( angle12 < -M_PI ) angle12 += 2.0*M_PI;
    if( angle12 != 0.0 )
      setTargetPos(FootQueue, 0.0, 0.0, angle12, nss, nds, verbose);
#else
    setTargetPos(FootQueue, middle_pos1(X), middle_pos1(Y), angle_pass, nss, nds, verbose);
#endif
    
    fsm->pushFootPlaceNoCheck(FootQueue, -1, step_before_down_stage, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, (double)nss*DT, (double)nds*DT);  
    fsm->pushFootPlaceNoCheck(FootQueue, -1, step_after_down_stage,  0.0, stage_height, 0.0, stage_height, 0.0, 0.0, 0.0, (double)nss*DT, (double)nds*DT);
    fsm->pushFootPlaceNoCheck(FootQueue, -1, step_after_down_stage,  0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, (double)nss*DT, (double)nds*DT);  

#if 1
    double angle2 = atan2(target_pos(Y)-middle_pos2(Y), target_pos(X)-middle_pos2(X));
    double length2 = sqrt((target_pos(X)-middle_pos2(X))*(target_pos(X)-middle_pos2(X)) + (target_pos(Y)-middle_pos2(Y))*(target_pos(Y)-middle_pos2(Y))); 
    if( angle2 != 0.0 )
      setTargetPos(FootQueue, 0.0, 0.0, angle2-angle_pass, nss, nds, verbose);
    if( length2 > 0.0 )
      setTargetPos(FootQueue, length2, 0.0, 0.0, nss, nds, verbose);
    
    double angle22 = target_pos(Z) - angle2;
    if( angle22 > M_PI ) angle22 -= 2.0*M_PI;
    else if( angle22 < -M_PI ) angle22 += 2.0*M_PI;
    if( angle22 != 0.0 )
      setTargetPos(FootQueue, 0.0, 0.0, angle22, nss, nds, verbose);
#else
    // notice: need to convert coordinates
    setTargetPos(FootQueue, target_pos(X)-middle_pos2(X), target_pos(Y)-middle_pos2(Y), target_pos(Z)-angle_pass, nss, nds, verbose);    
#endif    
  }
  else
    setTargetPos(FootQueue, target_pos(X), target_pos(Y), target_pos(Z), nss, nds, verbose);
  
  if( verbose ){
    for( int i = 0 ; i < (int)FootQueue.size() ; i++ )
      cout << "FootQueue(x,y,th) = " << FootQueue[i].p(X) << "," << FootQueue[i].p(Y) << "," << ToDeg(FootQueue[i].rpy(Z)) << ", nss = " << nss << ", nds = " << nds << endl;
  }
}
