/*
 * Copyright (c) 2009,
 * @author Mitsuharu Morisawa
 *
 * AIST
 *
 * All rights reserved.
 *
 * This program is made available under the terms of the Eclipse Public License
 * v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 */
#ifndef __PREVIEW_CONTROL_H__
#define __PREVIEW_CONTROL_H__

#include <ublasCommonTypes.h>
#include "LinearControl.h"

using namespace linear_control_system;
using namespace OpenHRP;


namespace linear_control_system{

  //#define ToVec(M) boost::numeric::ublas::matrix_row<dmatrix>(M,1)  

  class PreviewControl_base : virtual public LinearControl_SISO{
  public:
    int Nprev;
    int n_start;
    
    dmatrix F;    // preview control gain
    dvector ref;  // future trajectory
    
    PreviewControl_base(void);
    virtual ~PreviewControl_base();

    void Initialize(void);
    void setState(dvector& x_){x = x_;};
    void setSystemMatrices(dmatrix& A_, dmatrix& B_, dmatrix& C_){A = A_; B = B_; C = C_;};
    bool loadGains(const char *filename, int debug_level);
    void setReference(double dat, int n_now);
  };
  
  
  class PreviewControl : virtual public PreviewControl_base{
  public:
    PreviewControl(void);
    virtual ~PreviewControl();
    
    void reset(void);
    void update(double& p, double& v, double& a, double& y, int n_now);
    void update(double& p, double& v, double& a, int n_now);
  };
  
  
  class PreviewControlType1 : virtual public PreviewControl_base{
  private:
    double sx;
  public:
    PreviewControlType1(void);
    virtual ~PreviewControlType1();

    void reset(double sx = 0.0);
    void update(double& p, double& v, double& a, double& y, int n_now);    
    void update(double& p, double& v, double& a, int n_now);
  };
  
}; // end of namespace LinearControl

#endif // __PREVIEW_CONTROL_H__
