/*
 * Copyright (c) 2009,
 * @author Mitsuharu Morisawa
 *
 * AIST
 *
 * All rights reserved.
 *
 * This program is made available under the terms of the Eclipse Public License
 * v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 */
#include "ReadText.h"

using namespace read_text_parameter;


bool read_text_parameter::GetString(ifstream& in, string& str)
{
  char c;
  in.get(c);
  while( c == ' ' || c == '\t' || c== 0x0D || c == 0x0A || c == '#' ){
    if( !in ) return false;
    if( c == '#')
      while( c != 0x0A) in.get(c);
    in.get(c);
  }
  
  str.clear();
  while( c != ' ' && c != '\t' && c != 0x0D && c != 0x0A){
    str += c;
    in.get(c);
    if( !in ) return false;
  }
  
  in.putback(c);
  
  return true;
}
  
bool read_text_parameter::GetString( ifstream &in , char *str)
{
  char c;
  
  in.get(c);
  while( !in || c == ' ' || c == '\t' || c== 0x0D || c == '#' || c == 0x0A ){
    if( !in )
      return false;
    
    if( c == '#')
      while( c != 0x0A) in.get(c);
    in.get(c);
  }
  
  int i = 0;
  while( c != ' ' && c != '\t' && c != 0x0D && c != 0x0A){
    str[i++] = c;
    in.get(c);
    if( !in ){
      str[i] = 0;
      return false;
    }
  }
  str[i] = 0;
  in.putback(c);
  return true;
}

void read_text_parameter::GetVector(ifstream &fs, dvector& V)
{
  int m,n;
  GetData(fs, m);
  GetData(fs, n);
  V.resize(m*n);
  
  double x;
  for(int i = 0; i < m*n ; i++){
    GetData(fs, x);
    V(i) = x;
  }
}

void read_text_parameter::GetMatrix(ifstream &fs, dmatrix& M)
{
  int m,n;
  GetData(fs, m);
  GetData(fs, n);
  M.resize(m,n);
  
  double x;
  for(int i = 0; i < m; i++){
    for(int j = 0; j < n; j++){
      GetData(fs, x);
      M(i,j) = x;
    }
  }
}

void read_text_parameter::SkipToEOL( istream &in )
{
  char c;
  in.get(c);
  while( c != 0x0A && in) in.get(c); 
}
