/*
 * Copyright (c) 2009,
 * @author Mitsuharu Morisawa
 *
 * AIST
 *
 * All rights reserved.
 *
 * This program is made available under the terms of the Eclipse Public License
 * v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 */
#ifndef __MSTATE_LOGGER_H__
#define __MSTATE_LOGGER_H__

#include <queue>
#include <string>
#include "hrpModelHeaders.h"
#include "bodyinfo.h"
#include "mphysics.h"

using namespace OpenHRP;
using namespace std;


typedef struct{
  vector3 p;
  matrix33 R;
  vector3 v;
  vector3 w;
} waist_state;

typedef struct{
  vector3 p;
  matrix33 R;
  vector3 v;
  vector3 w;
} ankle_state;

typedef struct{
  vector3 p;
  vector3 v;
  vector3 rpy;
  vector3 drpy;
} foot_state;

class StateLogger{
  deque<waist_state> waist_queue;
  deque<vector<double> > pos_queue;
  deque<vector<double> > vel_queue;
  deque<vector3> zmp_queue;
  deque<vector3> hip_queue;
  deque<vector3> gsens_queue;
  deque<vector3> zmpref_queue;
  deque<vector3> pivot_queue[2];
  deque<ankle_state> ankle_queue[2];
  deque<foot_state> foot_queue[2];
  
 public:
  StateLogger(void);
  ~StateLogger();
  
  void clearLogData(void);
  void saveLogData(string& filename);
  
  void pushWaist(vector3& Pwaist, matrix33& Rwaist, vector3& Vwaist, vector3& Wwaist);
  void pushPos(vector<double>& pos){pos_queue.push_back(pos);};
  void pushVel(vector<double>& vel){vel_queue.push_back(vel);};
  void pushZMP(vector3& zmp){zmp_queue.push_back(zmp);};
  void pushHip(vector3& hip){hip_queue.push_back(hip);};
  void pushGsens(vector3& gsens){gsens_queue.push_back(gsens);};
  void pushZMPRef(vector3& zmpref){zmpref_queue.push_back(zmpref);};
  void pushPivot(vector3 *pivot);
  void pushAnkle(vector3& RPankle, matrix33& RRankle, vector3& RVankle, vector3& RWankle, vector3& LPankle, matrix33& LRankle, vector3& LVankle, vector3& LWankle);
  void pushFoot(vector3& RPfoot, vector3& RVfoot, vector3& RFoot_rpy, vector3& LPfoot, vector3& LVfoot, vector3& LFoot_rpy);
};

#endif // #ifndef __MSTATE_LOGGER_H__
