/*
 * Copyright (c) 2009,
 * @author Shuji Kajita, Shizuko Hattori, Mitsuharu Morisawa
 *
 * AIST
 *
 * All rights reserved.
 *
 * Redistribution in source file and binary forms, with or without modification is
 * strictly prohibited. Use in source and binary forms, with or without
 * modification is granted to the members the Joint AIST/CNRS Robotics Lab(JRL)
 * and the LAAS/CNRS lab.
 */
#ifndef __WHOLE_BODY_MOTION_GENERATOR_H__
#define __WHOLE_BODY_MOTION_GENERATOR_H__

#include "hrpModelHeaders.h"
#include "ModelManager.h"
#include "State.h"

using namespace OpenHRP;
using namespace std;

namespace motion_generator{
  
  #define USE_CHEST_RMC
  
  typedef tvmet::Vector<double, DOF>  Jvector;
  
  class WholeBodyMotionGenerator : public ModelManager{
    
    struct LinkInr{
      matrix33 Inr;
      vector3  Com;
      double   Mass;
      vector3  Hlin;
      vector3  Hang;
    };
    
#ifdef USE_CHEST_RMC
    const static int Nwbm = 6 + DOF_CHEST;
#else
    const static int Nwbm = 6;
#endif
    
  private:
    int       rootIndex;
    LinkInr   linkInr[DOF+1]; 
    
    vector3   total_com;
    matrix33  total_inr;
    
    dmatrix66 Jankle[2];
    dvector6  PLconst;
    dvector6  PLfree;
    
    dmatrix   Mconst;
    dmatrix66 Mconst_leg[2];
    dmatrix66 Mconst_arm[2];
#ifndef USE_CHEST_RMC
    dmatrix   Mconst_chest;
#endif
    dmatrix   Mconst_neck;
    
  protected:
    inline void makeCrossMatrix(matrix33& m, vector3& v)
    {
      m = 0.0,  -v(2),  v(1),
	  v(2),  0.0,  -v(0),
	 -v(1),  v(0),  0.0;
    }
    void initCOGInertia(void);
    void initWaistToAnkleJacobian(void);
    void initInertiaMatrix(void);
    void calcCOGInertia(vector3 &CoM, matrix33 &Inr);
    void setWaistToAnkleJacobian(State& Waist, State *Ankle);
    void setInertiaMatrix(State& Waist);
    void calcConstraintMomentum(State& Waist, State *Ankle);
    void generateReferenceMomentum(vector3& COG_cmd, vector3& VCOG_cmd, State& Waist_cmd, vector3& Zmp_ref, dvector& PLref);
    void calcInverseMomentum(dvector& PL, dvector& Vctrl, vector3& Zmp_cmd, dvector& Vctrl_ref);
    
  public:
    WholeBodyMotionGenerator(BodyPtr body_);
    ~WholeBodyMotionGenerator();
    
    void makeWholeBodyMotion(State& Waist_cmd, vector3& COG_cmd, vector3& VCOG_cmd, State *Ankle_cmd, vector3& Zmp_cmd, dvector& Vctrl_ref);
    void makeWholeBodyMotion(State& Waist_cmd, Jvector& q_cmd, Jvector& qdot_cmd, vector3& COG_cmd, vector3& VCOG_cmd, State *Ankle_cmd, vector3& Zmp_cmd, dvector& Vctrl_ref);
    void makeWholeBodyMotion(dvector& Vctrl, dvector& PL_ref, State& Waist_cmd, vector3& COG_cmd, vector3& VCOG_cmd, State *Ankle_cmd, vector3& Zmp_cmd, dvector& Vctrl_ref);
    void makeWholeBodyMotion(dvector& Vctrl, dvector& PL_ref, State& Waist_cmd, Jvector& q_cmd, Jvector& qdot_cmd, vector3& COG_cmd, vector3& VCOG_cmd, State *Ankle_cmd, vector3& Zmp_cmd, dvector& Vctrl_ref);
    
    void updateWholeBodyMotion(State& Waist_cmd, dvector& Vctrl_cmd);
    void updateWholeBodyMotion(State& Waist_cmd, Jvector& q_cmd, Jvector& dq_cmd, dvector& Vctrl_cmd);
    
    void calcMomentum(dvector6& PL_cmd, State& Waist_cmd, dvector& q_cmd, dvector& qdot_cmd);
  };

}  // end of namespace motion_generator


#endif // __WHOLE_BODY_MOTION_GENERATOR_H__
