/*
 * Copyright (c) 2009,
 * @author Mitsuharu Morisawa
 *
 * AIST
 *
 * All rights reserved.
 *
 * This program is made available under the terms of the Eclipse Public License
 * v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 */
#include "mwalk_common.h"

using namespace mwalk_pattern_generator;


vector3 mwalk_pattern_generator::omegaFromRpyAngularVelocity(vector3& rpy, vector3& rpy_dot)
{
  double croll = cos(rpy(1));
  double sroll = sin(rpy(1));
  double cyaw  = cos(rpy(2));
  double syaw  = sin(rpy(2));
  
  matrix33 R;
  R =
    croll*cyaw,  -syaw, 0.0,
    croll*syaw,  cyaw,  0.0,
    -sroll,      0.0,   1.0;
  
  return vector3(R * rpy_dot);
}

vector3 mwalk_pattern_generator::rpyAngularVelocityFromOmega(vector3& rpy, vector3& omega)
{
  double croll = cos(rpy(1));
  double sroll = sin(rpy(1));
  double cyaw  = cos(rpy(2));
  double syaw  = sin(rpy(2));
  
  matrix33 R;
  R =
    croll*cyaw,  croll*syaw,  -sroll,
    -syaw,       cyaw,        0.0,
    0.0,         0.0,         1.0;
  
  return vector3(R * omega);
}

double mwalk_pattern_generator::saturation(const double data, const double ulimit, const double llimit)
{
  if( data > ulimit )
    return ulimit;
  else if( data < llimit )
    return llimit;
  return data;
}

double mwalk_pattern_generator::threshold(const double data, const double ulimit, const double llimit)
{
  if( data > ulimit )
    return data - ulimit;
  else if( data < llimit )
    return data - llimit;
  return 0.0;
}

double mwalk_pattern_generator::normalizedTrapezoidCurve(int n_now, int n_ini, int n_acc, int n_dec, int n_end)
{
  if( n_now < n_ini )
    return 0.0;
  else if( n_now <= n_acc )
    return interp_func3((double)((n_now-n_ini)) / (double)((n_acc-n_ini)));
  else if( n_now <= n_dec )
    return 1.0;
  else if( n_now <= n_end )
    return interp_func3((double)((n_end-n_now)) / (double)((n_end-n_dec)));
  
  return 0.0;
}

string mwalk_pattern_generator::getWalkingStateString(WalkingState ws_)
{
  string str_state;
  
  switch(ws_ & (FROM_CENTER | FROM_RIGHT | FROM_LEFT | TO_CENTER | TO_RIGHT | TO_LEFT | SINGLE_SUPPORT | DOUBLE_SUPPORT)){
  case STAND:            str_state = STRING_STAND; break;
  case CENTER_TO_LEFT:   str_state = STRING_CENTER_TO_LEFT; break;
  case CENTER_TO_RIGHT:  str_state = STRING_CENTER_TO_RIGHT; break;
  case LEFT_TO_RIGHT:    str_state = STRING_LEFT_TO_RIGHT; break;
  case RIGHT_TO_LEFT:    str_state = STRING_RIGHT_TO_LEFT; break;
  case LEFT_TO_CENTER:   str_state = STRING_LEFT_TO_CENTER; break;
  case RIGHT_TO_CENTER:  str_state = STRING_RIGHT_TO_CENTER; break;
  case LEFT_SUPPORT:     str_state = STRING_LEFT_SUPPORT; break;
  case RIGHT_SUPPORT:    str_state = STRING_RIGHT_SUPPORT; break;
  case KEEP_RIGHT:       str_state = STRING_KEEP_RIGHT; break;
  case KEEP_LEFT:        str_state = STRING_KEEP_LEFT; break;
  default:                         break;
  }
  
  switch(ws_ & (FROM_ANKLE | FROM_TOE | TO_ANKLE | TO_TOE)){
  case ANKLE_SUPPORT:    str_state += string(" : ") + STRING_ANKLE_SUPPORT; break;
  case ANKLE_TO_TOE:     str_state += string(" : ") + STRING_ANKLE_TO_TOE; break;
  case TOE_SUPPORT:      str_state += string(" : ") + STRING_TOE_SUPPORT; break;
  case TOE_TO_ANKLE:     str_state += string(" : ") + STRING_TOE_TO_ANKLE; break;
  default:                        break;
  }
  
  return str_state;
}
