/*
 * Copyright (c) 2009,
 * @author Mitsuharu Morisawa
 *
 * AIST
 *
 * All rights reserved.
 *
 * This program is made available under the terms of the Eclipse Public License
 * v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 */
#ifndef __MWALK_COMMON_H__
#define __MWALK_COMMON_H__

#include "hrpModelHeaders.h"
#include "bodyinfo.h"
#include "mphysics.h"
#include "ModelManager.h"
#include "mwalkstate.h"

using namespace OpenHRP;
using namespace tvmet;
using namespace std;


namespace mwalk_pattern_generator{
  
  const static double       DEFAULT_FOOT_DISTANCE = 0.19;
  const static double       DEFAULT_COG_HEIGHT = 0.81;
  const static unsigned int Nstep = 3;                            ///< number of future foot steps for gait planning  
#ifndef WITH_TOE_JOINT
  const static int          Ncog = Nstep*2+1;                     ///< number of walking phase without toe joint
#else
  const static int          Ncog = Nstep*4+3;                     ///< number of walking phase with toe joint
#endif

  typedef enum {
    INITCONF,   ///< calculate initial COG and foot position
    ZMPFIX,     ///< dfilter mode, does not generate any walking pattern
    WALKING     ///< generate walking pattern
  } MotionMode;
  
  typedef ublas::bounded_matrix<double, Ncog*2+6, Ncog*2+6, ublas::column_major> ANmatrix;
  typedef ublas::bounded_vector<double, Ncog*2+6> ANvector;
  
  typedef tvmet::Matrix<double, 2, 2> matrix22;
  typedef tvmet::Matrix<double, 2, 4> matrix24;
  typedef tvmet::Vector<double, 2>    vector2;
  typedef tvmet::Vector<double, 4>    vector4;
  typedef tvmet::Vector<double, DOF>  Jvector;
  
  inline matrix22 inverse22(const matrix22& M)
  {
    double det = M(0,0)*M(1,1) - M(0,1)*M(1,0);
    if(fabs(det) < numeric_limits<double>::epsilon()){
      cerr << "can't calculate 2x2 inverse matrix : det = " << det << endl;
      cerr << "mat = " << M << endl;
      return matrix22(tvmet::identity<matrix22>()/numeric_limits<double>::epsilon()); //inverse(M);
      //return matrix22(0.0); //inverse(M);
    }
    
    matrix22 Minv;
    Minv =
      M(1,1)/det, -M(0,1)/det,
      -M(1,0)/det,  M(0,0)/det;
    return Minv;
  };
  
  static const vector3 zmp_ankle(0.0, 0.0, 0.0);                     /// default ZMP poisition at single support phase(Left foot standard in Y-axis)
  
  static const int     Nss = 140;                                    /// default single support period
  static const int     Nds = 20;                                     /// default double support period
  static const double  Tchange_ratio = 0.5;                          /// ratio which accept to modify foot place at single support phase
#define Nchange(n) (int)((double)(n)*Tchange_ratio)
  static const double  TPhip_updown_ini_ratio[2] = {0.0, 0.0};       /// ⤵ѤȤγϻ֤γ: {=ñӴ+ξӴ, ==ñӴΤ}
  static const double  TPhip_updown_end_ratio[2] = {1.0, 1.0};       /// ⤵ѤȤνλ֤γ: {=ñӴ+ξӴ, ==ñӴΤ}
#define NPhip_updown_ini(n,m) (int)((double)(n)*TPhip_updown_ini_ratio[m])
#define NPhip_updown_end(n,m) (int)((double)(n)*(TPhip_updown_end_ratio[m]-TPhip_updown_ini_ratio[m]))
  
#ifdef WITH_TOE_JOINT
#if defined (HRP2NF1) || defined(HRP4L) || defined(HRP4)
  //static const vector3 zmp_toe(0.135, 0.0, 0.0);                    /// ñӴץ졼ȤZMP(YϺ­)
  static const vector3 zmp_toe(0.0, 0.0, 0.0);                       /// ñӴץ졼ȤZMP(YϺ­)
  // tested by toe 10[deg]
  //static const double  TPzmp_ankle_to_toe_ratio[2] = {0.2,  0.6};   /// ñӴˤؤZMPΰư֤γ(, λ)
  //static const double  TPzmp_toe_to_ankle_ratio[2] = {0.2,  0.6};   /// ñӴˤ褫ؤZMPΰư֤γ(, λ)
  static const double  TPzmp_ankle_to_toe_ratio[2] = {0.45, 0.55};   /// ñӴˤؤZMPΰư֤γ(, λ)
  static const double  TPzmp_toe_to_ankle_ratio[2] = {0.45, 0.55};   /// ñӴˤ褫ؤZMPΰư֤γ(, λ)
  static const double  TRtoe_sup_ratio[2] = {0.8,  1.00 };           /// ñӴˤʤ֤γ(,λ) 
  static const double  TRtoe_swg_ratio[2] = {0.30, 0.95 };           /// ñӴˤ֤᤹γ (,λ) 
#define NPzmp_ankle_to_toe(n,m) (int)((double)(n)*TPzmp_ankle_to_toe_ratio[m])
#define NPzmp_toe_to_ankle(n,m) (int)((double)(n)*TPzmp_toe_to_ankle_ratio[m])
#define NRtoe_sup(n,m) (int)((double)(n)*TRtoe_sup_ratio[m])
#define NRtoe_swg(n,m) (int)((double)(n)*TRtoe_swg_ratio[m])
#endif
#endif // WITH_TOE_JOINT

#define interp_func3(t) ((3.0 - 2.0*(t))*(t)*(t))
#define interp_func6(t) (((6.0*(t) - 15.0)*(t) + 10.0)*(t)*(t)*(t))
  
  vector3 omegaFromRpyAngularVelocity(vector3& rpy, vector3& rpy_dot);
  vector3 rpyAngularVelocityFromOmega(vector3& rpy, vector3& omega);
  double  saturation(const double data, const double ulimit, const double llimit);
  double  threshold(const double data, const double ulimit, const double llimit);
  double  normalizedTrapezoidCurve(int n_now, int n_ini, int n_acc, int n_dec, int n_end);
  string  getWalkingStateString(WalkingState ws_);
  
} // end of namespace mwalk_pattern_generator

#endif // __MWALK_COMMON_H__
