/*
 * Copyright (c) 2009,
 * @author Mitsuharu Morisawa
 *
 * AIST
 *
 * All rights reserved.
 *
 * This program is made available under the terms of the Eclipse Public License
 * v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 */
#ifndef __MWALK_STATE_H__
#define __MWALK_STATE_H__

namespace mwalk_pattern_generator{

#define FROM_CENTER    0x0001
#define FROM_LEFT      0x0002
#define FROM_RIGHT     0x0004
#define FROM_ANKLE     0x0010
#define FROM_TOE       0x0020
#define TO_CENTER      0x0100
#define TO_LEFT        0x0200
#define TO_RIGHT       0x0400
#define TO_ANKLE       0x1000
#define TO_TOE         0x2000
#define SINGLE_SUPPORT 0x4000
#define DOUBLE_SUPPORT 0x8000

#define STAND                        (FROM_CENTER | TO_CENTER | DOUBLE_SUPPORT)
#define CENTER_TO_RIGHT              (FROM_CENTER | TO_RIGHT  | DOUBLE_SUPPORT)
#define CENTER_TO_LEFT               (FROM_CENTER | TO_LEFT   | DOUBLE_SUPPORT)
#define RIGHT_SUPPORT                (FROM_RIGHT  | TO_RIGHT  | SINGLE_SUPPORT)
#define RIGHT_TO_LEFT                (FROM_RIGHT  | TO_LEFT   | DOUBLE_SUPPORT)
#define RIGHT_TO_CENTER              (FROM_RIGHT  | TO_CENTER | DOUBLE_SUPPORT)
#define LEFT_SUPPORT                 (FROM_LEFT   | TO_LEFT   | SINGLE_SUPPORT)
#define LEFT_TO_RIGHT                (FROM_LEFT   | TO_RIGHT  | DOUBLE_SUPPORT)
#define LEFT_TO_CENTER               (FROM_LEFT   | TO_CENTER | DOUBLE_SUPPORT)
#define KEEP_RIGHT                   (FROM_RIGHT  | TO_RIGHT  | DOUBLE_SUPPORT)
#define KEEP_LEFT                    (FROM_LEFT   | TO_LEFT   | DOUBLE_SUPPORT)

#define ANKLE_SUPPORT                (FROM_ANKLE | TO_ANKLE)
#define TOE_SUPPORT                  (FROM_TOE   | TO_TOE)
#define ANKLE_TO_TOE                 (FROM_ANKLE | TO_TOE)
#define TOE_TO_ANKLE                 (FROM_TOE   | TO_ANKLE)

#define STAND_ANKLE                  (STAND | ANKLE_SUPPORT)
#define STAND_ANKLE_TO_TOE           (STAND | ANKLE_TOE_TOE)
#define STAND_TOE                    (STAND | TOE_SUPPORT)
#define STAND_TOE_TO_ANKLE           (STAND | TOE_TO_ANKLE)
#define CENTER_TO_RIGHT_ANKLE        (CENTER_TO_RIGHT | ANKLE_SUPPORT)
#define CENTER_TO_RIGHT_ANKLE_TO_TOE (CENTER_TO_RIGHT | ANKLE_TO_TOE)
#define CENTER_TO_RIGHT_TOE          (CENTER_TO_RIGHT | TOE_SUPPORT)
#define CENTER_TO_RIGHT_TOE_TO_ANKLE (CENTER_TO_RIGHT | TOE_TO_ANKLE)
#define CENTER_TO_LEFT_ANKLE         (CENTER_TO_LEFT | ANKLE_SUPPORT)
#define CENTER_TO_LEFT_ANKLE_TO_TOE  (CENTER_TO_LEFT | ANKLE_TO_TOE)
#define CENTER_TO_LEFT_TOE           (CENTER_TO_LEFT | TOE_SUPPORT)
#define CENTER_TO_LEFT_TOE_TO_ANKLE  (CENTER_TO_LEFT | TOE_TO_ANKLE)
#define RIGHT_TO_LEFT_ANKLE          (RIGHT_TO_LEFT | ANKLE_SUPPORT)
#define RIGHT_TO_LEFT_ANKLE_TO_TOE   (RIGHT_TO_LEFT | ANKLE_TO_TOE)
#define RIGHT_TO_LEFT_TOE            (RIGHT_TO_LEFT | TOE_SUPPORT)
#define RIGHT_TO_LEFT_TOE_TO_ANKLE   (RIGHT_TO_LEFT | TOE_TO_ANKLE)
#define RIGHT_TO_CENTER_ANKLE        (RIGHT_TO_CENTER | ANKLE_SUPPORT)
#define RIGHT_TO_CENTER_ANKLE_TO_TOE (RIGHT_TO_CENTER | ANKLE_TO_TOE)
#define RIGHT_TO_CENTER_TOE          (RIGHT_TO_CENTER | TOE_SUPPORT)
#define RIGHT_TO_CENTER_TOE_TO_ANKLE (RIGHT_TO_CENTER | TOE_TO_ANKLE)
#define LEFT_TO_RIGHT_ANKLE          (LEFT_TO_RIGHT | ANKLE_SUPPORT)
#define LEFT_TO_RIGHT_ANKLE_TO_TOE   (LEFT_TO_RIGHT | ANKLE_TO_TOE)
#define LEFT_TO_RIGHT_TOE            (LEFT_TO_RIGHT | TOE_SUPPORT)
#define LEFT_TO_RIGHT_TOE_TO_ANKLE   (LEFT_TO_RIGHT | TOE_TO_ANKLE)
#define LEFT_TO_CENTER_ANKLE         (LEFT_TO_CENTER | ANKLE_SUPPORT)
#define LEFT_TO_CENTER_ANKLE_TO_TOE  (LEFT_TO_CENTER | ANKLE_TO_TOE)
#define LEFT_TO_CENTER_TOE           (LEFT_TO_CENTER | TOE_SUPPORT)
#define LEFT_TO_CENTER_TOE_TO_ANKLE  (LEFT_TO_CENTER | TOE_TO_ANKLE)
#define RIGHT_SUPPORT_ANKLE          (RIGHT_SUPPORT | ANKLE_SUPPORT)
#define RIGHT_SUPPORT_ANKLE_TO_TOE   (RIGHT_SUPPORT | ANKLE_TO_TOE)
#define RIGHT_SUPPORT_TOE            (RIGHT_SUPPORT | TOE_SUPPORT)
#define RIGHT_SUPPORT_TOE_TO_ANKLE   (RIGHT_SUPPORT | TOE_TO_ANKLE)
#define LEFT_SUPPORT_ANKLE           (LEFT_SUPPORT | ANKLE_SUPPORT)
#define LEFT_SUPPORT_ANKLE_TO_TOE    (LEFT_SUPPORT | ANKLE_TO_TOE)
#define LEFT_SUPPORT_TOE             (LEFT_SUPPORT | TOE_SUPPORT)
#define LEFT_SUPPORT_TOE_TO_ANKLE    (LEFT_SUPPORT | TOE_TO_ANKLE)
#define KEEP_RIGHT_ANKLE             (KEEP_RIGHT | ANKLE_SUPPORT)
#define KEEP_RIGHT_ANKLE_TO_TOE      (KEEP_RIGHT | ANKLE_TO_TOE)
#define KEEP_RIGHT_TOE               (KEEP_RIGHT | TOE_SUPPORT)
#define KEEP_RIGHT_TOE_TO_ANKLE      (KEEP_RIGHT | TOE_TO_ANKLE)
#define KEEP_LEFT_ANKLE              (KEEP_LEFT | ANKLE_SUPPORT)
#define KEEP_LEFT_ANKLE_TO_TOE       (KEEP_LEFT | ANKLE_TO_TOE)
#define KEEP_LEFT_TOE                (KEEP_LEFT | TOE_SUPPORT)
#define KEEP_LEFT_TOE_TO_ANKLE       (KEEP_LEFT | TOE_TO_ANKLE)

#define STRING_STAND            "stand"
#define STRING_CENTER_TO_LEFT   "center to left"
#define STRING_RIGHT_TO_LEFT    "right to left"
#define STRING_KEEP_RIGHT       "keep right"
#define STRING_CENTER_TO_RIGHT  "center to right"
#define STRING_LEFT_TO_RIGHT    "left to right"
#define STRING_KEEP_LEFT        "keep left"
#define STRING_LEFT_TO_CENTER   "left to center"
#define STRING_RIGHT_TO_CENTER  "right to center"
#define STRING_LEFT_SUPPORT     "left support"
#define STRING_RIGHT_SUPPORT    "right support"
#define STRING_ANKLE_SUPPORT    "ankle support"
#define STRING_ANKLE_TO_TOE     "ankle to toe"
#define STRING_TOE_SUPPORT      "toe support"
#define STRING_TOE_TO_ANKLE     "toe to ankle"

#define getSupportLeg(a)   ((a & TO_RIGHT) == TO_RIGHT ? RLEG : ((a & TO_LEFT) == TO_LEFT ? LLEG : -1))
#define isSingleSupport(a) ((a & SINGLE_SUPPORT) == SINGLE_SUPPORT)
#define isDoubleSupport(a) ((a & DOUBLE_SUPPORT) == DOUBLE_SUPPORT)
#define isLeftSupport(a)   ((a & LEFT_SUPPORT) == LEFT_SUPPORT)
#define isRightSupport(a)  ((a & RIGHT_SUPPORT) == RIGHT_SUPPORT)
#define isAnkleSupport(a)  ((a & ANKLE_SUPPORT) == ANKLE_SUPPORT)
#define isToeSupport(a)    ((a & TOE_SUPPORT) == TOE_SUPPORT)

#define WalkingState int

} // end of namespace mwalk_pattern_generator

#endif // __MWALK_STATE_H__
